#!/usr/bin/env python3

# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import RNS
import argparse
import time
import sys
import os
import io
import base64

from RNS._version import __version__
from RNS.vendor import umsgpack as mp
from RNS.vendor.configobj import ConfigObj
from RNS.vendor.validate import Validator
from RNS.Cryptography.Hashes import sha256
from RNS.Cryptography.Hashes import file_sha256

APP_NAME = "rns"
DEFAULT_ASPECTS = f"{APP_NAME}.id"
NO_MESSAGE = 0x01
NO_META    = 0x02

PRV_EXT      = "rid"
PUB_EXT      = "pub"
SIG_EXT      = "rsg"
MSG_EXT      = "rsm"
ENCRYPT_EXT  = "rfe"
CHUNK_BLOCKS = 1024*1024
ENC_CHUNK    = CHUNK_BLOCKS*RNS.Identity.AES256_BLOCKSIZE
DEC_CHUNK    = ENC_CHUNK + RNS.Cryptography.Token.TOKEN_OVERHEAD*2

RSG_HASHTYPES = ["sha256"]

R_OK                = 0
R_NO_SIG_FILE       = 1
R_NO_IDENTITY       = 2
R_NO_PUBKEY         = 3
R_NO_PRVKEY         = 4
R_NO_KEYS           = 5
R_NO_FILE           = 6
R_INVALID_FILE      = 7
R_INVALID_IDENTITY  = 8
R_INVALID_ASPECTS   = 9
R_INVALID_SIGNATURE = 10
R_FILE_EXISTS       = 11
R_DECRYPT_FAILED    = 12
R_INVALID_ARGS      = 250
R_SEQUENCE_ERROR    = 251
R_READ_ERROR        = 252
R_WRITE_ERROR       = 253
R_UNKNOWN_ERROR     = 254
R_INTERRUPTED       = 255

reticulum = None

def validate_args(args):
    ops = 0;
    for o in [args.encrypt, args.decrypt, args.validate, args.sign, args.sign_message]:
        if o: ops += 1
    if ops > 1: print("This utility currently only supports one of the encrypt, decrypt, sign or verify operations per invocation"); exit(1)

    g = 0;
    for a in [args.import_pub, args.import_prv, args.identity, args.generate]:
        if a: g += 1
    if g > 1: print("The -i, -g, -m and -M args are mutually exclusive"); exit(1)

    g = 0;
    for a in [args.base64, args.base32, args.base256, args.hex]:
        if a: g += 1
    if g > 1: print("The -b, -B, --hex and --base256 args are mutually exclusive"); exit(1)

    return True

def main():
    try:
        parser = argparse.ArgumentParser(description="Reticulum Identity & Encryption Utility")

        # Identity Resolution
        parser.add_argument("--config", metavar="path", action="store", default=None, help="path to alternative Reticulum config directory", type=str)
        parser.add_argument("-i", "--identity", metavar="rid", action="store", default=None, help="hexadecimal Reticulum identity or destination hash, or path to Identity file", type=str)
        parser.add_argument("-g", "--generate", metavar="path", action="store", default=None, help="generate a new Identity and save to path")
        parser.add_argument("-m", "--import-pub", dest="import_pub", metavar="rid", action="store", default=None, help="import public Reticulum identity in hex, base32 or base64 format, or from file", type=str)
        parser.add_argument("-M", "--import-prv", dest="import_prv", metavar="rid", action="store", default=None, help="import Reticulum identity in hex, base32 or base64 format, or from file", type=str)
        parser.add_argument("-x", "--export-pub", action="store_true", default=None, help="export public identity to hex, base32 or base64 format")
        parser.add_argument("-X", "--export-prv", action="store_true", default=None, help="export private identity to hex, base32 or base64 format, or to file")

        # Verbosity Control
        parser.add_argument("-v", "--verbose", action="count", default=0, help="increase verbosity")
        parser.add_argument("-q", "--quiet", action="count", default=0, help="decrease verbosity")

        # Operations
        parser.add_argument("-a", "--announce", metavar="aspects", action="store", nargs="?", const=DEFAULT_ASPECTS, default=None, help="announce a destination based on this Identity")
        parser.add_argument("-H", "--hash", metavar="aspects", action="store", default=None, help="show destination hashes for other aspects for this Identity")
        parser.add_argument("-d", "--decrypt", metavar="file", action="store", nargs="*", default=None, help="decrypt file")
        parser.add_argument("-e", "--encrypt", metavar="file", action="store", nargs="*", default=None, help="encrypt file")
        parser.add_argument("-V", "--validate", metavar="path", action="store", nargs="*", default=None, help="validate signature")
        parser.add_argument("-s", "--sign", metavar="path", action="store", nargs="*", default=None, help="sign file")
        parser.add_argument("-S", "--sign-message", metavar="text", action="store", nargs="?", const=NO_MESSAGE, default=None, help="create embedded signed message")
        parser.add_argument("-E", "--embed-meta", metavar="path", action="store", nargs="?", const=NO_META, default=None, help="embed metadata structure from file")
        parser.add_argument("--meta-spec", metavar="path", action="store", nargs="?", default=None, help="validate metadata for embedding with spec from file")
        parser.add_argument("--raw", action="store_true", default=False, help="sign raw input data instead of hashing first")

        # I/O Control
        parser.add_argument("-w", "--write", metavar="path", action="store", default=None, help="output file path", type=str)
        parser.add_argument("-r", "--read", metavar="path", action="store", default=None, help="input file path for operations with optional file input", type=str)
        parser.add_argument("-f", "--force", action="store_true", default=None, help="write output even if it overwrites existing files")
        parser.add_argument("-I", "--stdin", action="store_true", default=False, help=argparse.SUPPRESS) # help="read input from STDIN instead of file"
        parser.add_argument("-O", "--stdout", action="store_true", default=False, help=argparse.SUPPRESS) # help="write output to STDOUT instead of file"

        # Information Flow
        parser.add_argument("-R", "--request", action="store_true", default=False, help="request unknown Identities from the network")
        parser.add_argument("-N", "--no-cache", action="store_true", default=False, help="never used cached or network-sourced information")
        parser.add_argument("-t", action="store", metavar="seconds", type=float, help="identity request timeout before giving up", default=RNS.Transport.PATH_REQUEST_TIMEOUT)
        parser.add_argument("-p", "--print-identity", action="store_true", default=False, help="print identity info and exit")
        parser.add_argument("-P", "--print-private", action="store_true", default=False, help="allow displaying private keys")

        # Formatting Control
        parser.add_argument("-B", "--base32", action="store_true", default=False, help="Use base32-encoded input and output")
        parser.add_argument("-b", "--base64", action="store_true", default=False, help="Use base64-encoded input and output")
        parser.add_argument("-U", "--base256", action="store_true", default=False, help="Use base256-encoded input and output")
        parser.add_argument("-F", "--hex", action="store_true", default=False, help="Use hex-encoded input and output")
        parser.add_argument("--meta", action="store_true", default=False, help="Display RSM metadata if available")

        parser.add_argument("--version", action="version", version="rnid {version}".format(version=__version__))
        
        args = parser.parse_args()
        validate_args(args)

        op_requires_identity = (args.sign or args.sign_message or args.encrypt or args.decrypt or args.announce or args.write
                                or args.print_identity or args.print_identity or args.export_pub or args.export_prv)

        identity = get_operating_identity(args, allow_none=not op_requires_identity, no_cache=args.no_cache); op = False
        if not identity and op_requires_identity: print("Could not get working identity"); exit(R_NO_IDENTITY)
        if args.print_identity: print_identity_information(args, identity); op = True
        if args.export_pub: export_pub_identity(args, identity); op = True
        if args.export_prv: export_prv_identity(args, identity); op = True
        if args.hash: print_hash_information(args, identity or args.identity); op = True
        if args.announce: announce(args, identity); op = True
        if args.validate: validate(args, identity or args.identity); op = True
        if args.sign: sign(args, identity); op = True
        if args.sign_message: sign_message(args, identity); op = True
        if args.encrypt: encrypt(args, identity); op = True
        if args.decrypt: decrypt(args, identity); op = True
        if args.write: write_identity(args, identity); op = True
        if args.generate: op = True

        if not op: parser.print_help()

        exit(0)

    except KeyboardInterrupt: print(""); exit(R_INTERRUPTED)


#####################
# Reticulum Helpers #
#####################

def ensure_reticulum(args):
    global reticulum
    if not reticulum:
        targetloglevel = 4; verbosity = args.verbose; quietness = args.quiet
        if verbosity != 0 or quietness != 0: targetloglevel = targetloglevel+verbosity-quietness
        reticulum = RNS.Reticulum(configdir=args.config, loglevel=targetloglevel)
        RNS.compact_log_fmt = True
        if args.stdout: RNS.loglevel = -1


#################################
# Identity Loading & Resolution #
#################################

def get_operating_identity(args, allow_none=False, no_cache=False):
    global reticulum
    identity = None

    if args.generate:
        identity = RNS.Identity()
        if not args.force and os.path.isfile(args.generate):
            print("Identity file "+str(args.generate)+" already exists. Not overwriting.")
            exit(R_FILE_EXISTS)
        
        else:
            try: identity.to_file(args.generate); print(f"New identity {identity} written to {args.generate}")
            except Exception as e: print(f"An error ocurred while saving the generated Identity: {e}"); exit(R_WRITE_ERROR)

    elif args.identity:
        load_path = None
        try: load_path = os.path.expanduser(args.identity)
        except: pass

        # Attempt to load Identity from .rid file
        if load_path and os.path.isfile(load_path):
            try:
                identity = RNS.Identity.from_file(load_path)
                print(f"Loaded Identity {identity} from {load_path}")
                if not identity.get_private_key() or not identity.get_public_key():
                    raise SystemError("Missing key data in loaded identity")

            except Exception as e: print(f"Could not load Identity from specified file: {e}"); exit(R_INVALID_IDENTITY)

        elif no_cache:
            if allow_none: return None
            else: print("Could not resolve identity"); exit(R_NO_IDENTITY)

        # Attempt to recall Identity from hex-encoded hash
        elif len(args.identity) == RNS.Reticulum.TRUNCATED_HASHLENGTH//8*2:
            try:
                ensure_reticulum(args)
                requested_hash = bytes.fromhex(args.identity)
                identity = RNS.Identity.recall(requested_hash) or RNS.Identity.recall(requested_hash, from_identity_hash=True)

                if identity == None:
                    if allow_none and not args.request: return None
                    elif not args.request:
                        print("Could not recall Identity for "+RNS.prettyhexrep(requested_hash)+".")
                        print("You can query the network for unknown Identities with the -R option.")
                        exit(R_NO_IDENTITY)
                    
                    else:
                        id_destination_hash = RNS.Destination.hash_from_name_and_identity(DEFAULT_ASPECTS, requested_hash)
                        RNS.Transport.request_path(requested_hash)      # Acquire if this was a destination hash
                        RNS.Transport.request_path(id_destination_hash) # Acquire if this was an identity hash
                        
                        def spincheck():
                            received = RNS.Identity.recall(requested_hash) or RNS.Identity.recall(requested_hash, from_identity_hash=True)
                            return received != None
                        spin(spincheck, "Requesting unknown Identity for "+RNS.prettyhexrep(requested_hash), args.t)

                        if not spincheck():
                            print("Identity request timed out");
                            if not allow_none: exit(R_NO_IDENTITY)
                            else: return None
                        
                        else:
                            identity = RNS.Identity.recall(requested_hash) or RNS.Identity.recall(requested_hash, from_identity_hash=True)
                            print("Received Identity "+str(identity)+" for destination "+RNS.prettyhexrep(requested_hash)+" from the network")

                else:
                    ident_str = str(identity)
                    hash_str  = RNS.prettyhexrep(requested_hash)
                    if ident_str == hash_str: print(f"Recalled Identity {ident_str}")
                    else:                     print(f"Recalled Identity {ident_str} for destination {hash_str}")

                if identity and identity.hash: reticulum._retain_identity(identity.hash)

            except Exception as e: print(f"Invalid hexadecimal hash provided: {e}"); exit(R_INVALID_IDENTITY)

    elif args.import_pub or args.import_prv:
        prvsize = RNS.Identity.KEYSIZE//8
        pubsize = prvsize
        identity_bytes = None
        if args.import_pub:
            try:
                identity_bytes = None
                import_path = os.path.expanduser(args.import_pub)
                if os.path.isfile(import_path):
                    try:
                        with open(import_path, "rb") as fh: file_input = fh.read()
                        if file_input and len(file_input) == pubsize:
                            identity_bytes = file_input
                            print(f"Reticulum Identity imported from {import_path}")
                    except: pass

                if not identity_bytes:
                    if len(args.import_pub) == pubsize*2:
                        try:
                            identity_bytes = bytes.fromhex(args.import_pub)
                            print("Reticulum Identity imported from hex input")
                        except: pass

                if not identity_bytes:
                    try:
                        b32_decoded = base64.b32decode(args.import_pub)
                        if len(b32_decoded) == pubsize:
                            identity_bytes = b32_decoded
                            print("Reticulum Identity imported from base32 input")
                    except: pass

                if not identity_bytes:
                    try:
                        b64_decoded = base64.urlsafe_b64decode(args.import_pub)
                        if len(b64_decoded) == pubsize:
                            identity_bytes = b64_decoded
                            print("Reticulum Identity imported from base64 input")
                    except: pass

                if not identity_bytes: print("Could not decode specified data to a valid public Reticulum Identity"); exit(R_INVALID_IDENTITY)

            except Exception as e: print("Invalid identity data specified for private identity import: "+str(e)); exit(R_INVALID_IDENTITY)

        elif args.import_prv:
            try:
                identity_bytes = None
                import_path = os.path.expanduser(args.import_prv)
                if os.path.isfile(import_path):
                    try:
                        with open(import_path, "rb") as fh: file_input = fh.read()
                        if file_input and len(file_input) == prvsize:
                            identity_bytes = file_input
                            print(f"Reticulum Identity imported from {import_path}")
                    except: pass

                if not identity_bytes:
                    if len(args.import_prv) == prvsize*2:
                        try:
                            identity_bytes = bytes.fromhex(args.import_prv)
                            print("Reticulum Identity imported from hex input")
                        except: pass

                if not identity_bytes:
                    try:
                        b32_decoded = base64.b32decode(args.import_prv)
                        if len(b32_decoded) == prvsize:
                            identity_bytes = b32_decoded
                            print("Reticulum Identity imported from base32 input")
                    except: pass

                if not identity_bytes:
                    try:
                        b64_decoded = base64.urlsafe_b64decode(args.import_prv)
                        if len(b64_decoded) == prvsize:
                            identity_bytes = b64_decoded
                            print("Reticulum Identity imported from base64 input")
                    except: pass

                if not identity_bytes: print("Could not decode specified data to a valid private Reticulum Identity"); exit(R_INVALID_IDENTITY)

            except Exception as e: print("Invalid identity data specified for private identity import: "+str(e)); exit(R_INVALID_IDENTITY)

        if args.import_prv:
            try: identity = RNS.Identity.from_bytes(identity_bytes)
            except Exception as e: print("Could not create Reticulum identity from specified data: "+str(e)); exit(R_INVALID_IDENTITY)

        elif args.import_pub:
            try:
                identity = RNS.Identity(create_keys=False)
                identity.load_public_key(identity_bytes)
            except Exception as e: print("Could not create Reticulum identity from specified data: "+str(e)); exit(R_INVALID_IDENTITY)

    return identity


######################
# Network Operations #
######################

def announce(args, identity):
    try:
        ensure_reticulum(args)
        aspects = args.announce.split(".")
        if not len(aspects) > 1: print("Invalid destination aspects specified"); exit(R_INVALID_ASPECTS)
        else:
            app_name = aspects[0]; aspects = aspects[1:]
            if not identity.get_private_key(): print("Cannot announce this destination, since the private key is not held"); exit(R_NO_PRVKEY)
            else:
                destination = RNS.Destination(identity, RNS.Destination.IN, RNS.Destination.SINGLE, app_name, *aspects)
                print(f"Announcing {args.announce} destination {RNS.prettyhexrep(destination.hash)} for identity {identity}")
                destination.announce(); time.sleep(0.25)

    except Exception as e: print(f"An error ocurred while attempting to send the announce: {e}"); exit(R_UNKNOWN_ERROR)


########################################
# Canonical RSG Manipulation Functions #
########################################

def get_rsg_data(rsg):
    rsg_data = None
    if   type(rsg) == bytes:             rsg_data = rsg
    elif type(rsg) == io.BufferedReader: rsg_data = rsg.read()
    elif type(rsg) == str:
        try:                             rsg_data = base64.urlsafe_b64decode(rsg)
        except:                          pass
        try:                             rsg_data = base64.b32decode(rsg.strip(RSG_PADDING))
        except:                          pass
        try:                             rsg_data = bytes.fromhex(rsg.strip(RSG_PADDING))
        except:                          pass
        try:                             rsg_data = RNS.b256_to_bytes(rsg.strip(RSG_PADDING.decode("utf-8")))
        except:                          pass

    return rsg_data

def extract_signed_rsg_data(rsg):
    siglen   = RNS.Identity.SIGLENGTH//8
    rsg_data = get_rsg_data(rsg)
    envelope = rsg_data[siglen:]

    try: return mp.unpackb(envelope)
    except: return None

def get_rsg_hash(message):
    sha = None
    if   type(message) == bytes:             sha = sha256(message)
    elif type(message) == str:               sha = sha256(message.encode("utf-8"))
    elif type(message) == io.BufferedReader: sha = file_sha256(message)
    else:                                    raise TypeError(f"Invalid input type {type(message)} for rsg creation")

    if not sha: raise ValueError("Hash calculation for rsg signature input failed")
    return sha

def rsg_is_legacy_format(rsg):
    rsg_data = get_rsg_data(rsg)
    if not rsg_data: return False
    return True if len(rsg_data) == RNS.Identity.SIGLENGTH//8 else False

def validate_rsg(rsg, message=None, required_signer=None):
    if not message: raise ValueError(f"No message specified for rsg validation")
    if not type(required_signer) in [RNS.Identity, bytes, type(None)]: raise TypeError(f"Invalid required signer type {type(required_signer)}")
    
    if   type(required_signer) == RNS.Identity: required_signer_hash = required_signer.hash
    elif type(required_signer) == bytes:        required_signer_hash = required_signer
    else:                                       required_signer_hash = None

    siglen = RNS.Identity.SIGLENGTH//8
    rsg_data = get_rsg_data(rsg)
    rsg_hash = get_rsg_hash(message)

    if len(rsg_data) == siglen: raise ValueError(f"Cannot validate legacy rsg format")

    if not rsg_data: return False, None, None
    else:
        if len(rsg_data) < siglen+1: return False, None, None
        else:
            signing_identity = None
            signature        = rsg_data[:siglen]
            envelope         = rsg_data[siglen:]

            try: signed_data = mp.unpackb(envelope)
            except: return False, None, None

            if not "hashtype" in signed_data or not "hash" in signed_data: return False, None, None
            if not signed_data["hashtype"] in RSG_HASHTYPES:               return False, None, None
            if not "meta" in signed_data:                                  return False, None, None
            if not "signer" in signed_data["meta"]:                        return False, None, None
            if not "pubkey" in signed_data["meta"]:                        return False, None, None

            try:
                if type(required_signer) == RNS.Identity:
                    signing_identity = required_signer
                
                else:
                    signing_identity = RNS.Identity(create_keys=False)
                    signing_identity.load_public_key(signed_data["meta"]["pubkey"])

            except: return False, None, None

            if required_signer_hash == None: required_signer_hash = signing_identity.hash

            if not signing_identity:                                       return False, None, None
            if not signing_identity.hash == required_signer_hash:          return False, None, signing_identity
            if signed_data["hash"] != rsg_hash:                            return False, None, signing_identity
            else:
                if not signing_identity.validate(signature, envelope):     return False, signed_data, signing_identity
                else:                                                      return True,  signed_data, signing_identity

            return False, signed_data, signing_identity

def create_rsg(signer_identity, message, embed=False, meta=None, output="bin"):
    if not output in ["bin", "hex", "base32", "base256", "base64"]: raise TypeError(f"Invalid output format for rsg creation")
    if not type(signer_identity) == RNS.Identity:                   raise TypeError(f"{signer_identity} is not a Reticulum Identity")
    if not signer_identity.get_private_key():                       raise ValueError(f"{signer_identity} does not hold a private key")

    signed_data = { "hashtype": "sha256", "hash": get_rsg_hash(message),
                    "meta": { "signer": signer_identity.hash,
                              "pubkey": signer_identity.get_public_key(),
                              "note"  : None } } # TODO: Remove default note field in 1.2.9

    if embed:
        if type(message) == str: message = message.encode("utf-8")
        signed_data["message"] = message

    if meta and type(meta) == dict:
        for key in meta:
            if not key in signed_data["meta"]: signed_data["meta"][key] = meta[key]

    envelope  = mp.packb(signed_data)
    signature = signer_identity.sign(envelope)
    rsg_data  = signature+envelope

    if   output == "bin":     rsg = rsg_data
    elif output == "hex":     rsg = RNS.hexrep(rsg_data, delimit=False).encode("ascii")
    elif output == "base32":  rsg = base64.b32encode(rsg_data)
    elif output == "base64":  rsg = base64.urlsafe_b64encode(rsg_data)
    elif output == "base256": rsg = RNS.b256rep(rsg_data)
    else:                     return None

    return rsg

RSG_ASCII_HEADER = b"#### Start of rsg data "
RSG_ASCII_FOOTER = b" End of rsg data ####"
RSG_ASCII_ROW_WIDTH = 64
RSG_PADDING = b"="
def wrap_rsg(rsg):
    if type(rsg) == str: return wrap_rsg_str(rsg)
    def pad(chunk): return chunk+(RSG_ASCII_ROW_WIDTH-len(chunk))*RSG_PADDING
    header = RSG_ASCII_HEADER+b"#"*(RSG_ASCII_ROW_WIDTH-len(RSG_ASCII_HEADER))
    footer = b"#"*(RSG_ASCII_ROW_WIDTH-len(RSG_ASCII_FOOTER))+RSG_ASCII_FOOTER
    wrapped = header+b"\n"
    read = 0
    while len(rsg):
        chunk = rsg[:RSG_ASCII_ROW_WIDTH]
        if len(chunk) < RSG_ASCII_ROW_WIDTH: chunk = pad(chunk)
        wrapped += chunk+b"\n"; read += len(chunk)
        rsg = rsg[len(chunk):]

    wrapped += footer
    return wrapped.decode("ascii")

def wrap_rsg_str(rsg):
    def pad(chunk): return chunk+(RSG_ASCII_ROW_WIDTH-len(chunk))*RSG_PADDING.decode("utf-8")
    header = RSG_ASCII_HEADER.decode("utf-8")+"#"*(RSG_ASCII_ROW_WIDTH-len(RSG_ASCII_HEADER.decode("utf-8")))
    footer = "#"*(RSG_ASCII_ROW_WIDTH-len(RSG_ASCII_FOOTER.decode("utf-8")))+RSG_ASCII_FOOTER.decode("utf-8")
    wrapped = header+"\n"
    read = 0
    while len(rsg):
        chunk = rsg[:RSG_ASCII_ROW_WIDTH]
        if len(chunk) < RSG_ASCII_ROW_WIDTH: chunk = pad(chunk)
        wrapped += chunk+"\n"; read += len(chunk)
        rsg = rsg[len(chunk):]

    wrapped += footer
    return wrapped

def unwrap_rsg(wrapped_rsg):
    unwrapped = ""
    if type(wrapped_rsg) == bytes: wrapped_rsg = wrapped_rsg.decode("ascii")
    elif type(wrapped_rsg) == str: pass
    else: return None

    for line in wrapped_rsg.splitlines():
        if not line.strip():     continue
        if line.startswith("#"): continue
        unwrapped += line

    return unwrapped if unwrapped else None

def rsg_meta_from_file(path, spec_path=None):
    if spec_path: meta_spec = ConfigObj(spec_path)
    else:         meta_spec = None
    parsed = ConfigObj(path, configspec=meta_spec)

    if meta_spec:
        validation = parsed.validate(Validator())
        if not validation == True: raise ValueError("Metadata did not pass spec validation")

    return parsed.dict()

def rsg_meta_from_str(meta, spec=None):
    if spec: meta_spec = ConfigObj(spec.splitlines())
    else:    meta_spec = None
    parsed = ConfigObj(meta.splitlines(), configspec=meta_spec)

    if meta_spec:
        validation = parsed.validate(Validator())
        if not validation == True: raise ValueError("Metadata did not pass spec validation")

    return parsed.dict()

def check_release_rsm_structure(signed_data):
    release_meta = signed_data.get("meta", None)
    if not release_meta: return "No release metadata in manifest"
    release_name = release_meta.get("name", None)
    release_version = release_meta.get("version", None)
    release_origin = release_meta.get("origin", None)
    release_origin_path = release_meta.get("path", None)
    if not release_name or not release_version:                     return "Incomplete package data in manifest"
    if not release_origin or not release_origin_path:               return "Incomplete release origin data in manifest"
    if "/" in release_name or "/" in release_version:               return "Invalid data in release manifest"
    if len(release_origin) != RNS.Identity.TRUNCATED_HASHLENGTH//8: return "Invalid origin hash length in manifest"
    if not type(release_origin) == bytes:                           return "Invalid origin hash in manifest"
    return True

###################################
# Signing & Validation Operations #
###################################

def validate(args, identity, __recursive=False):
    if type(args.validate) == list:
        paths     = args.validate.copy()
        validated = 0
        for path in paths:
            args.validate = path
            code = validate(args, identity, __recursive=True)
            if code != 0: print(f"Sequence error on recursive signature validation"); exit(R_SEQUENCE_ERROR)
            else:         validated += 1

        if len(paths) != validated: print(f"Sequence error on recursive signature validation"); exit(R_SEQUENCE_ERROR)
        else:                       exit(R_OK)

    msg_ext                            = f".{MSG_EXT}"
    sig_ext                            = f".{SIG_EXT}"
    validate_path                      = os.path.expanduser(args.validate)
    path_is_msgfile                    = validate_path.lower().endswith(msg_ext)
    path_is_sigfile                    = validate_path.lower().endswith(sig_ext)
    if path_is_sigfile: signature_path = validate_path; file_path = validate_path[:-len(sig_ext)]
    else:               signature_path = f"{validate_path}{sig_ext}"; file_path = validate_path
    signature_exists                   = os.path.isfile(signature_path)
    file_exists                        = os.path.isfile(file_path)

    if path_is_msgfile:      return validate_message(args, identity, __recursive=__recursive)
    if not file_exists:      print(f"The validation target \"{file_path}\" does not exist"); exit(R_NO_FILE)
    if not signature_exists: print(f"No signature file exists for \"{file_path}\""); exit(R_NO_FILE)

    try:
        with open(signature_path, "rb") as fh: is_legacy_format = rsg_is_legacy_format(fh)
    except Exception as e: print(f"Could not detect rsg format: {e}"); exit(R_UNKNOWN_ERROR)

    if not is_legacy_format:
        try:
            with open(signature_path, "rb") as fh: rsg = fh.read()
        except Exception as e: print(f"Could not read rsg: {e}"); exit(R_READ_ERROR)

        if type(identity) == str:
            if not len(identity) == RNS.Reticulum.TRUNCATED_HASHLENGTH//8*2: print("Invalid identity hash length"); exit(R_INVALID_IDENTITY)
            try: identity = bytes.fromhex(identity)
            except Exception as e: print(f"Invalid identity hash: {e}"); exit(R_INVALID_IDENTITY)

        try:
            with open(file_path, "rb") as fh:
                try:
                    valid, signed_data, signing_identity = validate_rsg(rsg, message=fh, required_signer=identity)
                    identity_str = RNS.prettyhexrep(identity) if type(identity) == bytes else f"{identity}"
                    signer_description = f"\nThis file was NOT signed by {identity_str or signing_identity}" if identity else ""
                    if not valid: print(f"Invalid signature {signature_path} for file {file_path}{signer_description}"); exit(R_INVALID_SIGNATURE)
                    else:         print(f"Signature is valid, the file {file_path} was signed by {signing_identity}"); return exit(R_OK) if not __recursive else R_OK

                except Exception as e: print(f"Error while validating {signature_path}: {e}"); exit(R_UNKNOWN_ERROR)

        except Exception as e: print(f"Could not read {file_path}: {e}"); exit(R_READ_ERROR)

    else:
        if type(identity) != RNS.Identity: print(f"Cannot validate legacy rsg signatures without an explicit required identity"); exit(R_NO_IDENTITY)
        try:
            with open(signature_path, "rb") as fh: signature = fh.read()
        except Exception as e: print(f"Could not read signature: {e}"); exit(R_READ_ERROR)

        try:
            with open(file_path, "rb") as fh: valid = identity.validate(signature, fh.read())
            if not valid: print(f"Invalid signature {signature_path} for file {file_path}\nThis file was NOT signed by {identity}"); exit(R_INVALID_SIGNATURE)
            else:         print(f"Signature is valid, the file {file_path} was signed by {identity}"); exit(R_OK)
        
        except Exception as e: print(f"Could not validate signature: {e}"); exit(R_READ_ERROR)

def validate_message(args, identity, __recursive=False):
    msg_ext                            = f".{MSG_EXT}"
    validate_path                      = os.path.expanduser(args.validate)
    path_is_msgfile                    = validate_path.lower().endswith(msg_ext)
    if path_is_msgfile: signature_path = validate_path
    signature_exists                   = os.path.isfile(signature_path)

    if not signature_exists: print(f"The signature file \"{signature_path}\" does not exist"); exit(R_NO_FILE)

    try:
        with open(signature_path, "rb") as fh: rsg = fh.read()
    except Exception as e: print(f"Could not read rsg: {e}"); exit(R_READ_ERROR)

    if type(identity) == str:
        if not len(identity) == RNS.Reticulum.TRUNCATED_HASHLENGTH//8*2: print("Invalid identity hash length"); exit(R_INVALID_IDENTITY)
        try: identity = bytes.fromhex(identity)
        except Exception as e: print(f"Invalid identity hash: {e}"); exit(R_INVALID_IDENTITY)

    try:
        rsm_contents = extract_signed_rsg_data(rsg)
        if not "message" in rsm_contents: print(f"No embedded message in {signature_path}"); exit(R_INVALID_SIGNATURE)
        valid, signed_data, signing_identity = validate_rsg(rsg, message=rsm_contents["message"], required_signer=identity)
        identity_str = RNS.prettyhexrep(identity) if type(identity) == bytes else f"{identity}"
        signer_description = f"\nThe message was NOT signed by {identity_str or signing_identity}" if identity else ""
        if not valid: print(f"Invalid signature in {signature_path}{signer_description}"); exit(R_INVALID_SIGNATURE)
        else:
            if args.meta:
                print("RSM Metadata\n============\n")
                def recurse(entry, key, level=1):
                    try:
                        indent = "  "*level
                        if type(entry) == dict:
                            print(f"d{indent}{key}:")
                            for key in entry: recurse(entry[key], key, level=level+1)
                        else:
                            maxwidth = 64
                            etype = "u"
                            if   type(entry) == str:   etype = "s"
                            elif type(entry) == bytes: etype = "b"
                            elif type(entry) == list:  etype = "l"
                            elif type(entry) == dict:  etype = "d"
                            elif type(entry) == int:   etype = "i"
                            elif type(entry) == float: etype = "f"
                            elif entry == None:        etype = "N"
                            if key == "note" and entry == None: return # TODO: Remove this check in 1.2.9
                            if type(entry) == bytes: entry = RNS.hexrep(entry, delimit=False)
                            leadin = f"{etype}{indent}{key}="; leadln = len(leadin)
                            entry = f"{entry}"; chunk = entry[:maxwidth]; entry = entry[maxwidth:]
                            print(f"{leadin}{chunk}")
                            while len(entry): chunk = entry[:maxwidth]; entry = entry[maxwidth:]; print(f" "*leadln+chunk)
                    except: print(f"E{indent}{key}=<Decode Error>")

                meta = signed_data["meta"]
                for key in meta: entry = meta[key]; recurse(entry, key)
                print("\nValidation\n==========")

            c = ":" if not args.meta else ""
            f = " following" if not args.meta else ""
            print(f"\nSignature is valid, the{f} message was signed by {signing_identity}{c}\n")
            if args.meta: print("Message\n=======\n")
            print(signed_data["message"].decode("utf-8"))

            return exit(R_OK) if not __recursive else R_OK

    except Exception as e: print(f"Error while validating {signature_path}: {e}"); exit(R_UNKNOWN_ERROR)

def sign(args, identity, __recursive=False):
    if type(args.sign) == list:
        paths  = args.sign.copy()
        signed = 0
        for path in paths:
            args.sign = path
            code = sign(args, identity, __recursive=True)
            if code != 0: print(f"Sequence error on recursive signature creation"); exit(R_SEQUENCE_ERROR)
            else:         signed += 1

        if len(paths) != signed: print(f"Sequence error on recursive signature creation"); exit(R_SEQUENCE_ERROR)
        else:                    exit(R_OK)

    sig_ext          = f".{SIG_EXT}"
    sign_path        = os.path.expanduser(args.sign)
    rsg_path         = f"{sign_path}{sig_ext}"
    file_exists      = os.path.isfile(sign_path)
    signature_exists = os.path.isfile(rsg_path)

    if   args.base32:  output = "base32"
    elif args.base64:  output = "base64"
    elif args.base256: output = "base256"
    elif args.hex:     output = "hex"
    else:              output = "bin"

    if not identity.get_private_key(): print(f"Cannot sign \"{sign_path}\", the identity does not hold a private key"); exit(R_NO_PRVKEY)
    if not file_exists: print(f"The file \"{sign_path}\" does not exist"); exit(R_NO_FILE)
    if output == "bin" and signature_exists and not args.force:
        print(f"The signature file \"{rsg_path}\" already exists, not overwriting"); exit(R_FILE_EXISTS)

    try:
        if args.raw:
            with open(sign_path, "rb") as fh: data = fh.read()
            with open(rsg_path, "wb") as fh: fh.write(identity.sign(data))

        else:
            with open(sign_path, "rb") as in_file: rsg = create_rsg(identity, in_file, output=output)
            if not rsg: print(f"No signature created, not writing"); exit(R_UNKNOWN_ERROR)

            if output == "bin":
                with open(rsg_path, "wb") as out_file: out_file.write(rsg)

            elif output in ["base32", "base64", "base256", "hex"]: print(f"\n{wrap_rsg(rsg)}\n")
            else: print("No valid output format specified"); exit(R_INVALID_ARGS)

        print(f"Signed file {sign_path} with {identity}"); return exit(R_OK) if not __recursive else R_OK

    except Exception as e: print(f"Could not sign {sign_path}: {e}"); exit(R_UNKNOWN_ERROR)

def sign_message(args, identity):
    message = args.sign_message
    meta = None

    if   args.base32:  output = "base32"
    elif args.base64:  output = "base64"
    elif args.base256: output = "base256"
    elif args.hex:     output = "hex"
    else:              output = "bin"

    if output == "bin" and not args.write: print("No write path specified"); exit(R_INVALID_ARGS)
    if not identity: print(f"Cannot sign, no working identity available"); exit(R_NO_IDENTITY)
    if not identity.get_private_key(): print(f"Cannot sign, the identity does not hold a private key"); exit(R_NO_PRVKEY)

    if args.read:
        if message != NO_MESSAGE: print("Both an input file and command-line provided message was specified, aborting"); exit(R_INVALID_ARGS)
        sign_path = os.path.expanduser(args.read)
        if not os.path.isfile(sign_path): print(f"The file {sign_path} does not exist"); exit(R_NO_FILE)
        with open(sign_path, "r", encoding="utf-8") as fh: message = fh.read()

    if message == NO_MESSAGE: message = get_editor_content()
    if not message: print("No message specified"); exit(R_INVALID_ARGS)

    if args.embed_meta:
        meta_path = os.path.expanduser(args.embed_meta)
        meta_spec_path = meta_path+".spec" if not args.meta_spec else args.meta_spec
        if not os.path.isfile(meta_path): print(f"Metadata file {meta_path} does not exist"); exit(R_NO_FILE)
        if not os.path.isfile(meta_spec_path): meta_spec_path = None
        spec_info = f" using spec from {meta_spec_path}" if meta_spec_path else ""
        print(f"Embedding metadata from {meta_path}{spec_info}")

        try: meta = rsg_meta_from_file(meta_path, spec_path=meta_spec_path)
        except Exception as e: print(f"Could not load metadata from {meta_path}: {e}"); exit(R_UNKNOWN_ERROR)

    try:
        rsg = create_rsg(identity, message, embed=True, meta=meta, output=output)
        if not rsg: print(f"No signature created, not writing"); exit(R_UNKNOWN_ERROR)

        if output == "bin":
            sig_ext          = f".{MSG_EXT}"
            rsg_path         = os.path.expanduser(args.write)
            rsg_path         = f"{rsg_path}{sig_ext}" if not rsg_path.endswith(sig_ext) else rsg_path
            signature_exists = os.path.isfile(rsg_path)
            if signature_exists and not args.force: print(f"The signature file \"{rsg_path}\" already exists, not overwriting"); exit(R_FILE_EXISTS)
            with open(rsg_path, "wb") as out_file: out_file.write(rsg)
            print(f"Message signed with {identity} saved to {rsg_path}"); exit(R_OK)

        elif output in ["base32", "base64", "base256", "hex"]: print(f"\n{wrap_rsg(rsg)}\n")
        else: print("No valid output format specified"); exit(R_INVALID_ARGS)

        print(f"Message signed with {identity}"); exit(R_OK)

    except Exception as e: print(f"Could not sign message: {e}"); exit(R_UNKNOWN_ERROR)


######################################
# Encryption & Decryption Operations #
######################################

def encrypt(args, identity, __recursive=False):
    if type(args.encrypt) == list:
        paths     = args.encrypt.copy()
        encrypted = 0
        for path in paths:
            args.encrypt = path
            code = encrypt(args, identity, __recursive=True)
            if code != 0: print(f"Sequence error on recursive file encryption"); exit(R_SEQUENCE_ERROR)
            else:         encrypted += 1

        if len(paths) != encrypted: print(f"Sequence error on recursive file encryption"); exit(R_SEQUENCE_ERROR)
        else:                       exit(R_OK)

    enc_ext      = f".{ENCRYPT_EXT}"
    encrypt_path = os.path.expanduser(args.encrypt)
    rfe_path     = args.write if args.write else f"{encrypt_path}{enc_ext}"
    file_exists  = os.path.isfile(encrypt_path)
    rfe_exists   = os.path.isfile(rfe_path)

    if not identity: print(f"Cannot encrypt \"{encrypt_path}\", no identity specified"); exit(R_NO_IDENTITY)
    if not identity.get_public_key(): print(f"Cannot encrypt \"{encrypt_path}\", the identity does not hold a public key"); exit(R_NO_PUBKEY)
    if not file_exists: print(f"The file \"{encrypt_path}\" does not exist"); exit(R_NO_FILE)

    if rfe_exists and not args.force:
        print(f"The encryption output file \"{rfe_path}\" already exists, not overwriting"); exit(R_FILE_EXISTS)

    try:
        with open(encrypt_path, "rb") as input_fh:
            try:
                with open(rfe_path, "wb") as output_fh:
                    wrote = 0
                    data_remaining = True
                    while data_remaining:
                        chunk = input_fh.read(ENC_CHUNK)
                        if chunk: wrote += output_fh.write(identity.encrypt(chunk))
                        else: data_remaining = False
                        print(f"\rWrote {RNS.prettysize(wrote)} to {rfe_path}   ", end="")

            except Exception as e: print(f"\nError writing encrypted output to {rfe_path}: {e}"); exit(R_WRITE_ERROR)
    except Exception as e: print(f"\nError reading {encrypt_path} for encryption: {e}"); exit(R_WRITE_ERROR)

    print(f"\nFile {encrypt_path} encrypted for {identity} to {rfe_path}"); return exit(R_OK) if not __recursive else R_OK

def decrypt(args, identity, __recursive=False):
    if type(args.decrypt) == list:
        paths     = args.decrypt.copy()
        decrypted = 0
        for path in paths:
            args.decrypt = path
            code = decrypt(args, identity, __recursive=True)
            if code != 0: print(f"Sequence error on recursive file decryption"); exit(R_SEQUENCE_ERROR)
            else:         decrypted += 1

        if len(paths) != decrypted: print(f"Sequence error on recursive file decryption"); exit(R_SEQUENCE_ERROR)
        else:                       exit(R_OK)

    enc_ext      = f".{ENCRYPT_EXT}"
    rfe_path     = os.path.expanduser(args.decrypt)
    if not rfe_path.endswith(enc_ext): print(f"The file {rfe_path} does not appear to be a Reticulum encrypted file"); exit(R_INVALID_FILE)

    decrypt_path = os.path.expanduser(args.write) if args.write else f"{rfe_path[:-len(enc_ext)]}"
    if not decrypt_path: print(f"Invalid output filename"); exit(R_INVALID_FILE)

    rfe_exists   = os.path.isfile(rfe_path)
    file_exists  = os.path.isfile(decrypt_path)

    if not identity: print(f"Cannot decrypt \"{rfe_path}\", no identity specified"); exit(R_NO_IDENTITY)
    if not identity.get_private_key(): print(f"Cannot decrypt \"{rfe_path}\", the identity does not hold a private key"); exit(R_NO_PRVKEY)
    if not rfe_exists: print(f"The file \"{rfe_path}\" does not exist"); exit(R_NO_FILE)

    if file_exists and not args.force:
        print(f"The decryption output file \"{decrypt_path}\" already exists, not overwriting"); exit(R_FILE_EXISTS)

    try:
        with open(rfe_path, "rb") as input_fh:
            try:
                with open(decrypt_path, "wb") as output_fh:
                    data_remaining = True
                    wrote = 0
                    while data_remaining:
                        chunk = input_fh.read(DEC_CHUNK)
                        if chunk:
                            decrypted = identity.decrypt(chunk)
                            if not decrypted: print(f"The provided identity could not decrypt the file"); exit(R_DECRYPT_FAILED)
                            else:             wrote += output_fh.write(decrypted)
                            print(f"\rWrote {RNS.prettysize(wrote)} to {decrypt_path}   ", end="")
                        
                        else: data_remaining = False

            except Exception as e: print(f"\nError writing decrypted output to {decrypt_path}: {e}"); exit(R_WRITE_ERROR)
    except Exception as e: print(f"\nError reading {rfe_path} for decryption: {e}"); exit(R_WRITE_ERROR)

    print(f"\nFile {rfe_path} decrypted to {decrypt_path}"); return exit(R_OK) if not __recursive else R_OK


################
#  File Output #
################

def write_identity(args, identity):
    try:
        wp = os.path.expanduser(args.write)
        args.write = False
        if identity.get_private_key() and args.export_prv:
            if not os.path.isfile(wp) or args.force:
                identity.to_file(wp)
                print("Wrote private identity to "+str(wp))
            else: print("File "+str(wp)+" already exists, not overwriting"); exit(R_FILE_EXISTS)
        
        elif identity.get_public_key():
            if not wp.lower().endswith(f".{PUB_EXT}"): wp += f".{PUB_EXT}"
            if not os.path.isfile(wp) or args.force:
                identity.pub_to_file(wp)
                print("Wrote public identity to "+str(wp))
            else: print("File "+str(wp)+" already exists, not overwriting"); exit(R_FILE_EXISTS)

        else: print("Identity holds neither a public nor private key"); exit(R_NO_KEYS)
    except Exception as e: print("Error while writing imported identity to file: "+str(e)); exit(R_WRITE_ERROR)


###################
# Terminal Output #
###################

def print_identity_information(args, identity):
    print("Identity Hash : "+RNS.prettyhexrep(identity.hash))
    if   args.base64: print("Public Key    : "+base64.urlsafe_b64encode(identity.get_public_key()).decode("utf-8"))
    elif args.base32: print("Public Key    : "+base64.b32encode(identity.get_public_key()).decode("utf-8"))
    else:             print("Public Key    : "+RNS.hexrep(identity.get_public_key(), delimit=False))
    
    if identity.prv:
        if args.print_private:
            if   args.base64: print("Private Key   : "+base64.urlsafe_b64encode(identity.get_private_key()).decode("utf-8"))
            elif args.base32: print("Private Key   : "+base64.b32encode(identity.get_private_key()).decode("utf-8"))
            else:             print("Private Key   : "+RNS.hexrep(identity.get_private_key(), delimit=False))
        else:                 print("Private Key   : Hidden")

def print_hash_information(args, identity):
    try:
        hashlen = RNS.Reticulum.TRUNCATED_HASHLENGTH//8*2
        if not type(identity) in [RNS.Identity, str]: print("Invalid identity"); exit(R_INVALID_IDENTITY)
        if type(identity) == str and len(identity) != hashlen: print("Invalid identity hash length"); exit(R_INVALID_IDENTITY)

        aspects = args.hash.split(".")
        if not len(aspects) > 0: print("Invalid destination aspects specified"); exit(R_INVALID_ASPECTS)
        else:
            app_name = aspects[0]; aspects = aspects[1:]
            if type(identity) == RNS.Identity and not identity.get_public_key(): print("Identity does not hold a public key"); exit(R_NO_PUBKEY)
            else:
                if type(identity) == RNS.Identity:
                    identity_hash = identity.hash
                    destination = RNS.Destination(identity, RNS.Destination.OUT, RNS.Destination.SINGLE, app_name, *aspects)
                
                elif type(identity) == str:
                    destination = None
                    try: identity_hash = bytes.fromhex(identity)
                    except Exception as e: print(f"Invalid identity: {e}"); exit(R_INVALID_IDENTITY)

                else: print("Invalid identity"); exit(R_INVALID_IDENTITY)

                destination_hash = RNS.Destination.hash_from_name_and_identity(args.hash, identity_hash)
                print(f"The {args.hash} destination for this Identity is {RNS.prettyhexrep(destination_hash)}")
                if destination: print("The full destination specifier is "+str(destination))
    
    except Exception as e: print(f"An error ocurred while attempting to get hash information: {e}"); exit(R_UNKNOWN_ERROR)

def export_pub_identity(args, identity):
    k = identity.get_public_key()
    if not k: print("Identity doesn't hold a public key, cannot export"); exit(R_NO_PUBKEY)
    else:
        if   args.base64: print("Public Identity Keys  : "+base64.urlsafe_b64encode(k).decode("utf-8"))
        elif args.base32: print("Public Identity Keys  : "+base64.b32encode(k).decode("utf-8"))
        else:             print("Public Identity Keys  : "+RNS.hexrep(k, delimit=False))

def export_prv_identity(args, identity):
    k = identity.get_private_key()
    if not k: print("Identity doesn't hold a private key, cannot export"); exit(R_NO_PRVKEY)
    else:
        if   args.base64: print("Private Identity Keys : "+base64.urlsafe_b64encode(k).decode("utf-8"))
        elif args.base32: print("Private Identity Keys : "+base64.b32encode(k).decode("utf-8"))
        else:             print("Private Identity Keys : "+RNS.hexrep(k, delimit=False))


##############################
# Helper & Utility Functions #
##############################

def get_editor_content():
    import subprocess
    from tempfile import NamedTemporaryFile
    template = ""
    editor = os.environ.get("EDITOR", "")
    if not editor:
        for fallback in ["nano", "vim", "vi"]:
            try:
                subprocess.run(["which", fallback], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                editor = fallback
                break
            except subprocess.CalledProcessError: continue

    if not editor: print("Could not launch editor"); exit(R_READ_ERROR);
    try:
        with NamedTemporaryFile(mode="w+", suffix=".tmp", delete=False) as tmp:
            tmp_path = tmp.name
            tmp.write(template)

        result = subprocess.run([editor, tmp_path])
        if result.returncode != 0: print(f"Editor exited with error code {result.returncode}"); os.unlink(tmp_path); exit(R_READ_ERROR)
        with open(tmp_path, "r") as f: content = f.read()
        os.unlink(tmp_path)
        return content.encode("utf-8")

    except Exception as e: print(f"Could not get content from editor: {e}"); exit(R_READ_ERROR)

def spin(until=None, msg=None, timeout=None):
    i = 0
    syms = "⢄⢂⢁⡁⡈⡐⡠"
    if timeout != None: timeout = time.time()+timeout

    print(msg+"  ", end=" ")
    while (timeout == None or time.time()<timeout) and not until():
        time.sleep(0.1)
        print(("\b\b"+syms[i]+" "), end="")
        sys.stdout.flush()
        i = (i+1)%len(syms)

    print("\r"+" "*len(msg)+"  \r", end="")

    if timeout != None and time.time() > timeout: return False
    else:                                         return True

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import RNS
import os
import sys
import time
import argparse

from RNS._version import __version__

remote_link = None
output_rst_str = "\r                                                          \r"
def connect_remote(destination_hash, auth_identity, timeout, no_output = False, purpose="management"):
    global remote_link, reticulum
    if not RNS.Transport.has_path(destination_hash):
        if not no_output:
            print("Path to "+RNS.prettyhexrep(destination_hash)+" requested", end=" ")
            sys.stdout.flush()
        RNS.Transport.request_path(destination_hash)
        pr_time = time.time()
        while not RNS.Transport.has_path(destination_hash):
            time.sleep(0.1)
            if time.time() - pr_time > timeout:
                if not no_output:
                    print(output_rst_str, end="")
                    print("Path request timed out")
                    exit(12)

    remote_identity = RNS.Identity.recall(destination_hash)

    def remote_link_closed(link):
        if link.teardown_reason == RNS.Link.INITIATOR_CLOSED: return
        elif link.teardown_reason == RNS.Link.TIMEOUT:
            if not no_output:
                print(output_rst_str, end="")
                print("The link timed out, exiting now")
        elif link.teardown_reason == RNS.Link.DESTINATION_CLOSED:
            if not no_output:
                print(output_rst_str, end="")
                print("The link was closed by the server, exiting now")
        else:
            if not no_output:
                print(output_rst_str, end="")
                print("Link closed unexpectedly, exiting now")
        exit(10)

    def remote_link_established(link):
        global remote_link
        if purpose == "management": link.identify(auth_identity)
        remote_link = link

    if not no_output:
        print(output_rst_str, end="")
        print("Establishing link with remote transport instance...", end=" ")
        sys.stdout.flush()

    if purpose == "management": remote_destination = RNS.Destination(remote_identity, RNS.Destination.OUT, RNS.Destination.SINGLE, "rnstransport", "remote", "management")
    elif purpose == "blackhole": remote_destination = RNS.Destination(remote_identity, RNS.Destination.OUT, RNS.Destination.SINGLE, "rnstransport", "info", "blackhole")
    link = RNS.Link(remote_destination)
    link.set_link_established_callback(remote_link_established)
    link.set_link_closed_callback(remote_link_closed)

def parse_hash(input_str):
    dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
    if len(input_str) != dest_len: raise ValueError("Hash length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
    try:
        hash_bytes = bytes.fromhex(input_str)
        return hash_bytes
    except Exception as e: raise ValueError("Invalid hash entered. Check your input.")

def program_setup(configdir, table, rates, drop, destination_hexhash, verbosity, timeout, drop_queues,
                  drop_via, max_hops, remote=None, management_identity=None, remote_timeout=RNS.Transport.PATH_REQUEST_TIMEOUT,
                  blackholed=False, blackhole=False, unblackhole=False, blackhole_duration=None, blackhole_reason=None,
                  remote_blackhole_list=False, remote_blackhole_list_filter=None, no_output=False, json=False):

    global remote_link, reticulum
    reticulum = RNS.Reticulum(configdir = configdir, loglevel = 3+verbosity)
    if remote:
        try:
            dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
            if len(remote) != dest_len: raise ValueError("Destination length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
            try:
                identity_hash = bytes.fromhex(remote)
                remote_hash = RNS.Destination.hash_from_name_and_identity("rnstransport.remote.management", identity_hash)
            except Exception as e: raise ValueError("Invalid destination entered. Check your input.")

            identity = RNS.Identity.from_file(os.path.expanduser(management_identity))
            if identity == None: raise ValueError("Could not load management identity from "+str(management_identity))

            try: connect_remote(remote_hash, identity, remote_timeout, no_output)
            except Exception as e: raise e

        except Exception as e:
            print(str(e))
            exit(20)

        while remote_link == None: time.sleep(0.1)

    if blackholed or remote_blackhole_list:
        blackholed_list = None
        if blackholed:
            if remote_link:
                if not no_output:
                    print(output_rst_str, end="")
                    print("Listing blackholed identities on remote instances not yet implemented")
                exit(255)

            try: blackholed_list = reticulum.get_blackholed_identities()
            except Exception as e:
                print(f"Could not get blackholed identities from RNS instance: {e}")
                exit(20)

        elif remote_blackhole_list:
            try: identity_hash = parse_hash(destination_hexhash)
            except Exception as e:
                print(f"{e}")
                exit(20)

            remote_hash = RNS.Destination.hash_from_name_and_identity("rnstransport.info.blackhole", identity_hash)
            connect_remote(remote_hash, None, remote_timeout, no_output, purpose="blackhole")
            while remote_link == None: time.sleep(0.1)

            if not no_output:
                print(output_rst_str, end="")
                print("Sending request...", end=" ")
                sys.stdout.flush()
            receipt = remote_link.request("/list")
            while not receipt.concluded(): time.sleep(0.1)
            response = receipt.get_response()
            if type(response) == dict:
                blackholed_list = response
                print(output_rst_str, end="")
            else:
                if not no_output:
                    print(output_rst_str, end="")
                    print("The remote request failed.")
                exit(10)

        else:
            print(f"Nowhere to fetch blackhole list from")
            exit(255)

        if not blackholed_list:
            print("No blackholed identity data available")
            exit(20)

        else:
            rmlen = 64
            def trunc(input_str):
                if len(input_str) <= rmlen: return input_str
                else: return f"{input_str[:rmlen-1]}…"

            try:
                now = time.time()
                for identity_hash in blackholed_list:
                    until      = blackholed_list[identity_hash]["until"]
                    reason     = blackholed_list[identity_hash]["reason"]
                    source     = blackholed_list[identity_hash]["source"]
                    until_str  = f"for {RNS.prettytime(max(0, until-now))}" if until else "indefinitely"
                    reason_str = f" ({trunc(reason)})" if reason else ""
                    by_str     = f" by {RNS.prettyhexrep(source)}" if source != RNS.Transport.identity.hash else ""
                    filter_str = f"{RNS.prettyhexrep(identity_hash)} {until_str} {reason_str} {by_str}"

                    if not remote_blackhole_list:
                        if destination_hexhash and not destination_hexhash in filter_str: continue
                    else:
                        if remote_blackhole_list_filter and not remote_blackhole_list_filter in filter_str: continue

                    print(f"{RNS.prettyhexrep(identity_hash)} blackholed {until_str}{reason_str}{by_str}")

            except Exception as e:
                print(f"Error while displaying collected blackhole data: {e}")
                exit(20)

    elif blackhole:
        if remote_link:
            if not no_output:
                print(output_rst_str, end="")
                print("Blackholing identity on remote instances not yet implemented")
            exit(255)

        try:
            identity_hash = parse_hash(destination_hexhash)
            until = time.time()+blackhole_duration*60*60 if blackhole_duration else None
            result = reticulum.blackhole_identity(identity_hash, until=until, reason=blackhole_reason)
            if result == True: print(f"Blackholed identity {destination_hexhash}")
            elif result == None: print(f"Identity {destination_hexhash} already blackholed")
            else: print(f"Could not blackhole identity {destination_hexhash}")
        
        except Exception as e:
            print(f"Could not blackhole identity: {e}")
            exit(20)
    
    elif unblackhole:
        if remote_link:
            if not no_output:
                print(output_rst_str, end="")
                print("Blackholing identity on remote instances not yet implemented")
            exit(255)

        try:
            identity_hash = parse_hash(destination_hexhash)
            result = reticulum.unblackhole_identity(identity_hash)
            if result == True: print(f"Lifted blackhole for identity {destination_hexhash}")
            elif result == None: print(f"Identity {destination_hexhash} not blackholed")
            else: print(f"Could not unblackhole identity {destination_hexhash}")
        
        except Exception as e:
            print(f"Could not unblackhole identity: {e}")
            exit(20)
    
    elif table:
        destination_hash = None
        if destination_hexhash != None:
            try:
                dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
                if len(destination_hexhash) != dest_len: raise ValueError("Destination length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
                try: destination_hash = bytes.fromhex(destination_hexhash)
                except Exception as e: raise ValueError("Invalid destination entered. Check your input.")
            except Exception as e:
                print(str(e))
                sys.exit(1)

        if not remote_link: table = sorted(reticulum.get_path_table(max_hops=max_hops), key=lambda e: (e["interface"], e["hops"]) )
        else:
            if not no_output:
                print(output_rst_str, end="")
                print("Sending request...", end=" ")
                sys.stdout.flush()
            receipt = remote_link.request("/path", data = ["table", destination_hash, max_hops])
            while not receipt.concluded(): time.sleep(0.1)
            response = receipt.get_response()
            if response:
                table = response
                print(output_rst_str, end="")
            else:
                if not no_output:
                    print(output_rst_str, end="")
                    print("The remote request failed. Likely authentication failure.")
                exit(10)

        displayed = 0
        if json:
            import json
            for p in table:
                for k in p:
                    if isinstance(p[k], bytes): p[k] = RNS.hexrep(p[k], delimit=False)

            print(json.dumps(table))
            exit()
        
        else:
            for path in table:
                if destination_hash == None or destination_hash == path["hash"]:
                    displayed += 1
                    exp_str = RNS.timestamp_str(path["expires"])
                    if path["hops"] == 1: m_str = " "
                    else: m_str = "s"
                    print(RNS.prettyhexrep(path["hash"])+" is "+str(path["hops"])+" hop"+m_str+" away via "+RNS.prettyhexrep(path["via"])+" on "+path["interface"]+" expires "+RNS.timestamp_str(path["expires"]))

            if destination_hash != None and displayed == 0:
                print("No path known")
                sys.exit(1)

    elif rates:
        destination_hash = None
        if destination_hexhash != None:
            try:
                dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
                if len(destination_hexhash) != dest_len: raise ValueError("Destination length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
                try: destination_hash = bytes.fromhex(destination_hexhash)
                except Exception as e: raise ValueError("Invalid destination entered. Check your input.")
            except Exception as e:
                print(str(e))
                sys.exit(1)

        if not remote_link: table = reticulum.get_rate_table()
        else:
            if not no_output:
                print(output_rst_str, end="")
                print("Sending request...", end=" ")
                sys.stdout.flush()
            receipt = remote_link.request("/path", data = ["rates", destination_hash])
            while not receipt.concluded():
                time.sleep(0.1)
            response = receipt.get_response()
            if response:
                table = response
                print(output_rst_str, end="")
            else:
                if not no_output:
                    print(output_rst_str, end="")
                    print("The remote request failed. Likely authentication failure.")
                exit(10)

        table = sorted(table, key=lambda e: e["last"])
        if json:
            import json
            for p in table:
                for k in p:
                    if isinstance(p[k], bytes): p[k] = RNS.hexrep(p[k], delimit=False)

            print(json.dumps(table))
            exit()
        else:
            if len(table) == 0: print("No information available")
            else:
                displayed = 0
                for entry in table:
                    if destination_hash == None or destination_hash == entry["hash"]:
                        displayed += 1
                        try:
                            last_str = pretty_date(int(entry["last"]))
                            start_ts = entry["timestamps"][0]
                            span = max(time.time() - start_ts, 3600.0)
                            span_hours = span/3600.0
                            span_str = pretty_date(int(entry["timestamps"][0]))
                            hour_rate = round(len(entry["timestamps"])/span_hours, 3)
                            if hour_rate-int(hour_rate) == 0:
                                hour_rate = int(hour_rate)
                            
                            if entry["rate_violations"] > 0:
                                if entry["rate_violations"] == 1:
                                    s_str = ""
                                else:
                                    s_str = "s"

                                rv_str = ", "+str(entry["rate_violations"])+" active rate violation"+s_str
                            else:
                                rv_str = ""
                            
                            if entry["blocked_until"] > time.time():
                                bli = time.time()-(int(entry["blocked_until"])-time.time())
                                bl_str = ", new announces allowed in "+pretty_date(int(bli))
                            else:
                                bl_str = ""

            
                            print(RNS.prettyhexrep(entry["hash"])+" last heard "+last_str+" ago, "+str(hour_rate)+" announces/hour in the last "+span_str+rv_str+bl_str)

                        except Exception as e:
                            print("Error while processing entry for "+RNS.prettyhexrep(entry["hash"]))
                            print(str(e))

                if destination_hash != None and displayed == 0:
                    print("No information available")
                    sys.exit(1)

    elif drop_queues:
        if remote_link:
            if not no_output:
                print(output_rst_str, end="")
                print("Dropping announce queues on remote instances not yet implemented")
            exit(255)

        print("Dropping announce queues on all interfaces...")
        reticulum.drop_announce_queues()
    
    elif drop:
        if remote_link:
            if not no_output:
                print(output_rst_str, end="")
                print("Dropping path on remote instances not yet implemented")
            exit(255)

        try:
            dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
            if len(destination_hexhash) != dest_len: raise ValueError("Destination length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
            try: destination_hash = bytes.fromhex(destination_hexhash)
            except Exception as e: raise ValueError("Invalid destination entered. Check your input.")
        except Exception as e:
            print(str(e))
            sys.exit(1)

        if reticulum.drop_path(destination_hash): print("Dropped path to "+RNS.prettyhexrep(destination_hash))
        else:
            print("Unable to drop path to "+RNS.prettyhexrep(destination_hash)+". Does it exist?")
            sys.exit(1)

    elif drop_via:
        if remote_link:
            if not no_output:
                print(output_rst_str, end="")
                print("Dropping all paths via specific transport instance on remote instances yet not implemented")
            exit(255)

        try:
            dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
            if len(destination_hexhash) != dest_len: raise ValueError("Destination length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
            try: destination_hash = bytes.fromhex(destination_hexhash)
            except Exception as e: raise ValueError("Invalid destination entered. Check your input.")
        except Exception as e:
            print(str(e))
            sys.exit(1)

        if reticulum.drop_all_via(destination_hash): print("Dropped all paths via "+RNS.prettyhexrep(destination_hash))
        else:
            print("Unable to drop paths via "+RNS.prettyhexrep(destination_hash)+". Does the transport instance exist?")
            sys.exit(1)

    else:
        if remote_link:
            if not no_output:
                print(output_rst_str, end="")
                print("Requesting paths on remote instances not implemented")
            exit(255)

        try:
            dest_len = (RNS.Reticulum.TRUNCATED_HASHLENGTH//8)*2
            if len(destination_hexhash) != dest_len: raise ValueError("Destination length is invalid, must be {hex} hexadecimal characters ({byte} bytes).".format(hex=dest_len, byte=dest_len//2))
            try: destination_hash = bytes.fromhex(destination_hexhash)
            except Exception as e: raise ValueError("Invalid destination entered. Check your input.")
        except Exception as e:
            print(str(e))
            sys.exit(1)
            
        if not RNS.Transport.has_path(destination_hash):
            RNS.Transport.request_path(destination_hash)
            print("Path to "+RNS.prettyhexrep(destination_hash)+" requested  ", end=" ")
            sys.stdout.flush()

        i = 0
        syms = "⢄⢂⢁⡁⡈⡐⡠"
        limit = time.time()+timeout
        while not RNS.Transport.has_path(destination_hash) and time.time()<limit:
            time.sleep(0.1)
            print(("\b\b"+syms[i]+" "), end="")
            sys.stdout.flush()
            i = (i+1)%len(syms)

        if RNS.Transport.has_path(destination_hash):
            hops = RNS.Transport.hops_to(destination_hash)
            next_hop_bytes = reticulum.get_next_hop(destination_hash)
            if next_hop_bytes == None:
                print("\r                                                       \rError: Invalid path data returned")
                sys.exit(1)
            else:
                next_hop = RNS.prettyhexrep(next_hop_bytes)
                next_hop_interface = reticulum.get_next_hop_if_name(destination_hash)

                if hops != 1: ms = "s"
                else: ms = ""

                print("\rPath found, destination "+RNS.prettyhexrep(destination_hash)+" is "+str(hops)+" hop"+ms+" away via "+next_hop+" on "+next_hop_interface)
        else:
            print("\r                                                       \rPath not found")
            sys.exit(1)


def main():
    try:
        parser = argparse.ArgumentParser(description="Reticulum Path Management Utility")
        parser.add_argument("--config", action="store", default=None, help="path to alternative Reticulum config directory", type=str)
        parser.add_argument("--version", action="version", version="rnpath {version}".format(version=__version__))
        parser.add_argument("-t", "--table", action="store_true", help="show all known paths", default=False)
        parser.add_argument("-m", "--max", action="store", metavar="hops", type=int, help="maximum hops to filter path table by", default=None)
        parser.add_argument("-r", "--rates", action="store_true", help="show announce rate info", default=False)
        parser.add_argument("-d", "--drop", action="store_true", help="remove the path to a destination", default=False)
        parser.add_argument("-D", "--drop-announces", action="store_true", help="drop all queued announces", default=False)
        parser.add_argument("-x", "--drop-via", action="store_true", help="drop all paths via specified transport instance", default=False)
        parser.add_argument("-w", action="store", metavar="seconds", type=float, help="timeout before giving up", default=RNS.Transport.PATH_REQUEST_TIMEOUT)
        parser.add_argument("-R", action="store", metavar="hash", help="transport identity hash of remote instance to manage", default=None, type=str)
        parser.add_argument("-i", action="store", metavar="path", help="path to identity used for remote management", default=None, type=str)
        parser.add_argument("-W", action="store", metavar="seconds", type=float, help="timeout before giving up on remote queries", default=RNS.Transport.PATH_REQUEST_TIMEOUT)
        parser.add_argument("-b", "--blackholed", action="store_true", help="list blackholed identities", default=False)
        parser.add_argument("-B", "--blackhole", action="store_true", help="blackhole identity", default=False)
        parser.add_argument("-U", "--unblackhole", action="store_true", help="unblackhole identity", default=False)
        parser.add_argument(      "--duration", action="store", type=float, help="duration of blackhole enforcement in hours", default=None)
        parser.add_argument(      "--reason", action="store", type=str, help="reason for blackholing identity", default=None)
        parser.add_argument("-p", "--blackholed-list", action="store_true", help="view published blackhole list for remote transport instance", default=False)
        parser.add_argument("-j", "--json", action="store_true", help="output in JSON format", default=False)
        parser.add_argument("destination", nargs="?", default=None, help="hexadecimal hash of the destination", type=str)
        parser.add_argument("list_filter", nargs="?", default=None, help="filter for remote blackhole list view", type=str)
        parser.add_argument('-v', '--verbose', action='count', default=0)
        
        args = parser.parse_args()

        if args.config: configarg = args.config
        else: configarg = None

        if not args.drop_announces and not args.table and not args.rates and not args.destination and not args.drop_via and not args.blackholed:
            print("")
            parser.print_help()
            print("")
        else:
            program_setup(configdir = configarg, table = args.table, rates = args.rates, drop = args.drop, destination_hexhash = args.destination,
                          verbosity = args.verbose, timeout = args.w, drop_queues = args.drop_announces, drop_via = args.drop_via, max_hops = args.max,
                          remote=args.R, management_identity=args.i, remote_timeout=args.W, blackholed=args.blackholed, blackhole=args.blackhole,
                          unblackhole=args.unblackhole, blackhole_duration=args.duration, blackhole_reason=args.reason, remote_blackhole_list=args.blackholed_list,
                          remote_blackhole_list_filter=args.list_filter, json=args.json)

            sys.exit(0)

    except KeyboardInterrupt:
        print("")
        exit()

def pretty_date(time=False):
    from datetime import datetime
    now = datetime.now()
    if type(time) is int: diff = now - datetime.fromtimestamp(time)
    elif isinstance(time,datetime): diff = now - time
    elif not time: diff = now - now
    second_diff = diff.seconds
    day_diff = diff.days
    if day_diff < 0: return ''
    if day_diff == 0:
        if second_diff < 10: return str(second_diff) + " seconds"
        if second_diff < 60: return str(second_diff) + " seconds"
        if second_diff < 70: return "1 minute"
        if second_diff < 7200: return str(int(second_diff / 60)) + " minutes"
        if second_diff < 86400: return str(int(second_diff / 3600)) + " hours"
    if day_diff == 1: return "1 day"
    if day_diff < 7: return str(day_diff) + " days"
    if day_diff < 31: return str(int(day_diff / 7)) + " weeks"
    if day_diff < 365: return str(int(day_diff / 30)) + " months"
    return str(int(day_diff / 365)) + " years"

if __name__ == "__main__": main()


# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import glob

from .Hashes import sha256
from .Hashes import sha512
from .HKDF import hkdf
from .PKCS7 import PKCS7
from .Token import Token
from .Provider import backend

import RNS.Cryptography.Provider as cp

if cp.PROVIDER == cp.PROVIDER_INTERNAL:
    from RNS.Cryptography.X25519 import X25519PrivateKey, X25519PublicKey
    from RNS.Cryptography.Ed25519 import Ed25519PrivateKey, Ed25519PublicKey

elif cp.PROVIDER == cp.PROVIDER_PYCA:
    from RNS.Cryptography.Proxies import X25519PrivateKeyProxy as X25519PrivateKey
    from RNS.Cryptography.Proxies import X25519PublicKeyProxy as X25519PublicKey
    from RNS.Cryptography.Proxies import Ed25519PrivateKeyProxy as Ed25519PrivateKey
    from RNS.Cryptography.Proxies import Ed25519PublicKeyProxy as Ed25519PublicKey

py_modules  = glob.glob(os.path.dirname(__file__)+"/*.py")
pyc_modules = glob.glob(os.path.dirname(__file__)+"/*.pyc")
modules     = py_modules+pyc_modules
__all__ = list(set([os.path.basename(f).replace(".pyc", "").replace(".py", "") for f in modules if not (f.endswith("__init__.py") or f.endswith("__init__.pyc"))]))


# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
import os
import RNS
import time
import atexit
import hashlib
import threading

from .vendor import umsgpack as umsgpack

from RNS.Cryptography import X25519PrivateKey, X25519PublicKey, Ed25519PrivateKey, Ed25519PublicKey
from RNS.Cryptography import Token


class Identity:
    """
    This class is used to manage identities in Reticulum. It provides methods
    for encryption, decryption, signatures and verification, and is the basis
    for all encrypted communication over Reticulum networks.

    :param create_keys: Specifies whether new encryption and signing keys should be generated.
    """

    CURVE = "Curve25519"
    """
    The curve used for Elliptic Curve DH key exchanges
    """

    KEYSIZE     = 256*2
    """
    X.25519 key size in bits. A complete key is the concatenation of a 256 bit encryption key, and a 256 bit signing key.
    """

    RATCHETSIZE = 256
    """
    X.25519 ratchet key size in bits.
    """

    RATCHET_EXPIRY = 60*60*24*30
    """
    The expiry time for received ratchets in seconds, defaults to 30 days. Reticulum will always use the most recently
    announced ratchet, and remember it for up to ``RATCHET_EXPIRY`` since receiving it, after which it will be discarded.
    If a newer ratchet is announced in the meantime, it will be replace the already known ratchet.
    """

    # Non-configurable constants
    TOKEN_OVERHEAD            = RNS.Cryptography.Token.TOKEN_OVERHEAD
    AES128_BLOCKSIZE          = 16          # In bytes
    AES256_BLOCKSIZE          = 16          # In bytes
    HASHLENGTH                = 256         # In bits
    SIGLENGTH                 = KEYSIZE     # In bits

    NAME_HASH_LENGTH          = 80
    TRUNCATED_HASHLENGTH      = RNS.Reticulum.TRUNCATED_HASHLENGTH
    """
    Constant specifying the truncated hash length (in bits) used by Reticulum
    for addressable hashes and other purposes. Non-configurable.
    """

    DERIVED_KEY_LENGTH        = 512//8
    DERIVED_KEY_LENGTH_LEGACY = 256//8

    # Storage
    known_destinations = {}
    known_ratchets = {}

    ratchet_persist_lock = threading.Lock()
    known_destinations_lock = threading.Lock()

    @staticmethod
    def remember(packet_hash, destination_hash, public_key, app_data = None):
        if len(public_key) != Identity.KEYSIZE//8:
            raise TypeError("Can't remember "+RNS.prettyhexrep(destination_hash)+", the public key size of "+str(len(public_key))+" is not valid.", RNS.LOG_ERROR)
        else:
            with Identity.known_destinations_lock:
                if not destination_hash in Identity.known_destinations:
                    Identity.known_destinations[destination_hash] = [time.time(), packet_hash, public_key, app_data, 0]
                else:
                    entry = Identity.known_destinations[destination_hash]
                    entry[0] = time.time()
                    entry[1] = packet_hash
                    entry[2] = public_key
                    entry[3] = app_data

    @staticmethod
    def recall(target_hash, from_identity_hash=False, _no_use=False):
        """
        Recall identity for a destination or identity hash. By default, this function
        will return the identity associated with a given *destination* hash. As an
        example, if you know the ``lxmf.delivery`` destination hash of an endpoint,
        this function will return the associated underlying identity. You can also
        search for an identity from a known *identity hash*, by setting the
        ``from_identity_hash`` argument.

        :param target_hash: Destination or identity hash as *bytes*.
        :param from_identity_hash: Whether to search based on identity hash instead of destination hash as *bool*.
        :returns: An :ref:`RNS.Identity<api-identity>` instance that can be used to create an outgoing :ref:`RNS.Destination<api-destination>`, or *None* if the destination is unknown.
        """
        if from_identity_hash:
            with Identity.known_destinations_lock: destination_hashes = list(Identity.known_destinations.keys())
            for destination_hash in destination_hashes:
                entry = Identity.known_destinations.get(destination_hash)
                if not entry: continue
                if target_hash == Identity.truncated_hash(entry[2]):
                    if not _no_use: RNS.Reticulum.get_instance()._used_destination_data(destination_hash)
                    identity = Identity(create_keys=False)
                    identity.load_public_key(entry[2])
                    identity.app_data = entry[3]
                    return identity

            return None

        else:
            if target_hash in Identity.known_destinations:
                if not _no_use: RNS.Reticulum.get_instance()._used_destination_data(target_hash)
                identity_data = Identity.known_destinations[target_hash]
                identity = Identity(create_keys=False)
                identity.load_public_key(identity_data[2])
                identity.app_data = identity_data[3]
                return identity
            else:
                for registered_destination in RNS.Transport.destinations:
                    if target_hash == registered_destination.hash:
                        identity = Identity(create_keys=False)
                        identity.load_public_key(registered_destination.identity.get_public_key())
                        identity.app_data = None
                        return identity

                return None

    @staticmethod
    def recall_app_data(destination_hash, _no_use=False):
        """
        Recall last heard app_data for a destination hash.

        :param destination_hash: Destination hash as *bytes*.
        :returns: *Bytes* containing app_data, or *None* if the destination is unknown.
        """
        if destination_hash in Identity.known_destinations:
            if not _no_use: RNS.Reticulum.get_instance()._used_destination_data(destination_hash)
            app_data = Identity.known_destinations[destination_hash][3]
            return app_data
        
        else: return None

    @staticmethod
    def save_known_destinations(background=False, recombine=True):
        # TODO: Improve the storage method so we don't have to
        # deserialize and serialize the entire table on every
        # save, but the only changes. It might be possible to
        # simply overwrite on exit now that every local client
        # disconnect triggers a data persist.
        
        try:
            if hasattr(Identity, "saving_known_destinations"):
                wait_interval = 0.2
                wait_timeout = 5
                wait_start = time.time()
                while Identity.saving_known_destinations:
                    time.sleep(wait_interval)
                    if time.time() > wait_start+wait_timeout:
                        RNS.log("Could not save known destinations to storage, waiting for previous save operation timed out.", RNS.LOG_ERROR)
                        return False

            Identity.saving_known_destinations = True
            save_start = time.time()

            if recombine:
                storage_known_destinations = {}
                if os.path.isfile(RNS.Reticulum.storagepath+"/known_destinations"):
                    try:
                        with open(RNS.Reticulum.storagepath+"/known_destinations","rb") as file:
                            storage_known_destinations = umsgpack.load(file)
     
                    except: pass

                try:
                    for destination_hash in storage_known_destinations:
                        if not destination_hash in Identity.known_destinations:
                            with Identity.known_destinations_lock:
                                Identity.known_destinations[destination_hash] = storage_known_destinations[destination_hash]
                
                except Exception as e:
                    RNS.log("Skipped recombining known destinations from disk, since an error occurred: "+str(e), RNS.LOG_WARNING)

            RNS.log("Saving "+str(len(Identity.known_destinations))+" known destinations to storage...", RNS.LOG_VERBOSE)
            temp_file = RNS.Reticulum.storagepath+f"/known_destinations.tmp.{time.time()}"

            try:
                with open(temp_file,"wb") as file: umsgpack.dump(Identity.known_destinations.copy(), file)
                os.rename(temp_file, RNS.Reticulum.storagepath+f"/known_destinations")

            except Exception as e:
                RNS.log(f"Error while serializing and writing known destinations: {e}", RNS.LOG_ERROR)
                try: os.unlink(temp_file)
                except Exception as e: RNS.log(f"Could not clean up temporary file {temp_file}: {e}", RNS.LOG_WARNING)
                raise e

            save_time = time.time() - save_start
            if save_time < 1: time_str = str(round(save_time*1000,2))+"ms"
            else:             time_str = str(round(save_time,2))+"s"

            RNS.log("Saved known destinations to storage in "+time_str, RNS.LOG_VERBOSE)

        except Exception as e:
            RNS.log("Error while saving known destinations to disk, the contained exception was: "+str(e), RNS.LOG_ERROR)
            RNS.trace_exception(e)

        Identity.saving_known_destinations = False

    @staticmethod
    def load_known_destinations():
        if os.path.isfile(RNS.Reticulum.storagepath+"/known_destinations"):
            st = time.time()
            try:
                with open(RNS.Reticulum.storagepath+"/known_destinations","rb") as file:
                    loaded_known_destinations = umsgpack.load(file)

                Identity.known_destinations = {}
                for known_destination in loaded_known_destinations:
                    if len(known_destination) == RNS.Reticulum.TRUNCATED_HASHLENGTH//8:
                        if len(loaded_known_destinations[known_destination]) < 5:
                            e = loaded_known_destinations[known_destination]
                            loaded_known_destinations[known_destination] = [e[0], e[1], e[2], e[3], 0]

                        with Identity.known_destinations_lock:
                            Identity.known_destinations[known_destination] = loaded_known_destinations[known_destination]

                RNS.log(f"Loaded {len(Identity.known_destinations)} known destination from storage in {RNS.prettyshorttime(time.time()-st)}", RNS.LOG_VERBOSE)

            except Exception as e:
                RNS.log("Error loading known destinations from disk, file will be recreated on exit", RNS.LOG_ERROR)
                RNS.trace_exception(e)
        else:
            RNS.log("Destinations file does not exist, no known destinations loaded", RNS.LOG_VERBOSE)

    @staticmethod
    def _used_destination_data(destination_hash):
        with Identity.known_destinations_lock:
            if destination_hash in Identity.known_destinations:
                if not Identity.known_destinations[destination_hash][4] < 0:
                    Identity.known_destinations[destination_hash][4] = time.time()
                    return True

        return False

    @staticmethod
    def _retain_destination_data(destination_hash):
        with Identity.known_destinations_lock:
            if destination_hash in Identity.known_destinations:
                Identity.known_destinations[destination_hash][4] = -1
                return True

        return False

    @staticmethod
    def _unretain_destination_data(destination_hash):
        with Identity.known_destinations_lock:
            if destination_hash in Identity.known_destinations:
                Identity.known_destinations[destination_hash][4] = time.time()
                return True

        return False

    @staticmethod
    def _retain_identity(identity_hash):
        try:
            retained = False
            with Identity.known_destinations_lock: destination_hashes = list(Identity.known_destinations.keys())
            for destination_hash in destination_hashes:
                entry = Identity.known_destinations.get(destination_hash)
                if not entry: continue
                if identity_hash == Identity.truncated_hash(entry[2]):
                    if Identity._retain_destination_data(destination_hash): retained = True

            return retained

        except Exception as e: RNS.log(f"Error while retaining identity {RNS.prettyhexrep(identity_hash)}: {e}", RNS.LOG_ERROR)
    
    @staticmethod
    def clean_known_destinations():
        now        = time.time()
        st         = now
        total      = len(Identity.known_destinations)
        stale      = []
        no_path    = 0
        retained   = 0
        never_used = 0

        with Identity.known_destinations_lock: destination_hashes = list(Identity.known_destinations.keys())
        for destination_hash in destination_hashes:
            try:
                if RNS.Transport.has_path(destination_hash): has_path = True
                else:
                    has_path = False
                    no_path += 1

                with Identity.known_destinations_lock:
                    if destination_hash in Identity.known_destinations:
                        last_announce =  Identity.known_destinations[destination_hash][0]
                        last_use = 0
                        was_used = False
                        is_retained = False

                        if Identity.known_destinations[destination_hash][4] > 0:
                            was_used = True
                            last_use = Identity.known_destinations[destination_hash][4]

                        elif Identity.known_destinations[destination_hash][4] == 0:
                            was_used = False
                            never_used += 1

                        elif Identity.known_destinations[destination_hash][4] == -1:
                            is_retained = True
                            retained += 1

                        unused_for = time.time() - Identity.known_destinations[destination_hash][4]

                        if not is_retained and not has_path:
                            if not was_used and now - last_announce > RNS.Transport.UNUSED_DESTINATION_LINGER: stale.append(destination_hash)
                            elif unused_for > RNS.Transport.DESTINATION_TIMEOUT*1.25:                          stale.append(destination_hash)

            except Exception as e: RNS.log(f"Faulty entry for {RNS.prettyhexrep(destination_hash)} while cleaning known destinations: {e}", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None

        removed = 0
        for destination_hash in stale:
            with Identity.known_destinations_lock:
                if destination_hash in Identity.known_destinations:
                    Identity.known_destinations.pop(destination_hash)
                    removed += 1

        # RNS.log(f"Total destinations: {total}, stale: {len(stale)}, removed: {removed}, no path: {no_path}, never used: {never_used}, with path: {total-no_path}, used: {total-never_used}, retained: {retained}. Completed in {RNS.prettyshorttime(time.time()-st)}", RNS.LOG_WARNING) # TODO: Remove
        if not RNS.Transport.owner.is_connected_to_shared_instance: Identity.save_known_destinations(recombine=False)

    @staticmethod
    def full_hash(data):
        """
        Get a SHA-256 hash of passed data.

        :param data: Data to be hashed as *bytes*.
        :returns: SHA-256 hash as *bytes*.
        """
        return RNS.Cryptography.sha256(data)

    @staticmethod
    def truncated_hash(data):
        """
        Get a truncated SHA-256 hash of passed data.

        :param data: Data to be hashed as *bytes*.
        :returns: Truncated SHA-256 hash as *bytes*.
        """
        return Identity.full_hash(data)[:(Identity.TRUNCATED_HASHLENGTH//8)]

    @staticmethod
    def get_random_hash():
        """
        Get a random SHA-256 hash.

        :param data: Data to be hashed as *bytes*.
        :returns: Truncated SHA-256 hash of random data as *bytes*.
        """
        return Identity.truncated_hash(os.urandom(Identity.TRUNCATED_HASHLENGTH//8))

    @staticmethod
    def current_ratchet_id(destination_hash):
        """
        Get the ID of the currently used ratchet key for a given destination hash

        :param destination_hash: A destination hash as *bytes*.
        :returns: A ratchet ID as *bytes* or *None*.
        """
        ratchet = Identity.get_ratchet(destination_hash)
        if ratchet == None:
            return None
        else:
            return Identity._get_ratchet_id(ratchet)

    @staticmethod
    def _get_ratchet_id(ratchet_pub_bytes):
        return Identity.full_hash(ratchet_pub_bytes)[:Identity.NAME_HASH_LENGTH//8]

    @staticmethod
    def _ratchet_public_bytes(ratchet):
        return X25519PrivateKey.from_private_bytes(ratchet).public_key().public_bytes()

    @staticmethod
    def _generate_ratchet():
        ratchet_prv = X25519PrivateKey.generate()
        ratchet_pub = ratchet_prv.public_key()
        return ratchet_prv.private_bytes()

    @staticmethod
    def _remember_ratchet(destination_hash, ratchet):
        try:
            if destination_hash in Identity.known_ratchets and Identity.known_ratchets[destination_hash] == ratchet:
                ratchet_exists = True
            else:
                ratchet_exists = False

            if not ratchet_exists:
                RNS.log(f"Remembering ratchet {RNS.prettyhexrep(Identity._get_ratchet_id(ratchet))} for {RNS.prettyhexrep(destination_hash)}", RNS.LOG_EXTREME) if RNS.sl(RNS.LOG_EXTREME) else None
                Identity.known_ratchets[destination_hash] = ratchet
                if not RNS.Transport.owner.is_connected_to_shared_instance:
                    def persist_job():
                        with Identity.ratchet_persist_lock:
                            hexhash = RNS.hexrep(destination_hash, delimit=False)
                            ratchet_data = {"ratchet": ratchet, "received": time.time()}

                            ratchetdir = RNS.Reticulum.storagepath+"/ratchets"
                            
                            if not os.path.isdir(ratchetdir):
                                os.makedirs(ratchetdir)

                            outpath   = f"{ratchetdir}/{hexhash}.out"
                            finalpath = f"{ratchetdir}/{hexhash}"
                            with open(outpath, "wb") as ratchet_file:
                                ratchet_file.write(umsgpack.packb(ratchet_data))
                            os.replace(outpath, finalpath)

                    
                    threading.Thread(target=persist_job, daemon=True).start()

        except Exception as e:
            RNS.log(f"Could not persist ratchet for {RNS.prettyhexrep(destination_hash)} to storage.", RNS.LOG_ERROR)
            RNS.log(f"The contained exception was: {e}")
            RNS.trace_exception(e)

    @staticmethod
    def _clean_ratchets():
        RNS.log("Cleaning ratchets...", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
        try:
            count = 0
            removed = 0
            not_known = 0
            now = time.time()
            ratchetdir = RNS.Reticulum.storagepath+"/ratchets"
            if os.path.isdir(ratchetdir):
                for filename in os.listdir(ratchetdir):
                    count += 1
                    try:
                        expired = False
                        corrupted = False
                        with open(f"{ratchetdir}/{filename}", "rb") as rf:
                            try:
                                ratchet_data = umsgpack.unpackb(rf.read())
                                if now > ratchet_data["received"]+Identity.RATCHET_EXPIRY: expired = True

                            except Exception as e:
                                RNS.log(f"Corrupted ratchet data while reading {ratchetdir}/{filename}, removing file", RNS.LOG_ERROR)
                                corrupted = True

                        destination_hash = bytes.fromhex(filename)
                        if not destination_hash in RNS.Identity.known_destinations: unknown = True; not_known += 1
                        else:                                                       unknown = False

                        if expired or corrupted or unknown:
                            os.unlink(f"{ratchetdir}/{filename}")
                            removed += 1

                    except Exception as e:
                        RNS.log(f"An error occurred while cleaning ratchets, in the processing of {ratchetdir}/{filename}.", RNS.LOG_ERROR)
                        RNS.log(f"The contained exception was: {e}", RNS.LOG_ERROR)

        except Exception as e: RNS.log(f"An error occurred while cleaning ratchets. The contained exception was: {e}", RNS.LOG_ERROR)
        RNS.log(f"Processed {count} ratchets in {RNS.prettytime(time.time()-now)}, not in use {not_known}, removed {removed}", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None

    @staticmethod
    def get_ratchet(destination_hash):
        if not destination_hash in Identity.known_ratchets:
            ratchetdir = RNS.Reticulum.storagepath+"/ratchets"
            hexhash = RNS.hexrep(destination_hash, delimit=False)
            ratchet_path = f"{ratchetdir}/{hexhash}"
            if os.path.isfile(ratchet_path):
                try:
                    with open(ratchet_path, "rb") as ratchet_file:
                        ratchet_data = umsgpack.unpackb(ratchet_file.read())
                        if time.time() < ratchet_data["received"]+Identity.RATCHET_EXPIRY and len(ratchet_data["ratchet"]) == Identity.RATCHETSIZE//8:
                            Identity.known_ratchets[destination_hash] = ratchet_data["ratchet"]
                        else:
                            return None
                
                except Exception as e:
                    RNS.log(f"An error occurred while loading ratchet data for {RNS.prettyhexrep(destination_hash)} from storage.", RNS.LOG_ERROR)
                    RNS.log(f"The contained exception was: {e}", RNS.LOG_ERROR)
                    return None

        if destination_hash in Identity.known_ratchets:
            return Identity.known_ratchets[destination_hash]
        else:
            RNS.log(f"Could not load ratchet for {RNS.prettyhexrep(destination_hash)}", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
            return None

    @staticmethod
    def validate_announce(packet, only_validate_signature=False):
        try:
            if packet.packet_type == RNS.Packet.ANNOUNCE:
                keysize       = Identity.KEYSIZE//8
                ratchetsize   = Identity.RATCHETSIZE//8
                name_hash_len = Identity.NAME_HASH_LENGTH//8
                sig_len       = Identity.SIGLENGTH//8
                destination_hash = packet.destination_hash

                # Get public key bytes from announce
                public_key = packet.data[:keysize]

                # If the packet context flag is set,
                # this announce contains a new ratchet
                if packet.context_flag == RNS.Packet.FLAG_SET:
                    name_hash   = packet.data[keysize:keysize+name_hash_len ]
                    random_hash = packet.data[keysize+name_hash_len:keysize+name_hash_len+10]
                    ratchet     = packet.data[keysize+name_hash_len+10:keysize+name_hash_len+10+ratchetsize]
                    signature   = packet.data[keysize+name_hash_len+10+ratchetsize:keysize+name_hash_len+10+ratchetsize+sig_len]
                    app_data    = b""
                    if len(packet.data) > keysize+name_hash_len+10+sig_len+ratchetsize:
                        app_data = packet.data[keysize+name_hash_len+10+sig_len+ratchetsize:]

                # If the packet context flag is not set,
                # this announce does not contain a ratchet
                else:
                    ratchet     = b""
                    name_hash   = packet.data[keysize:keysize+name_hash_len]
                    random_hash = packet.data[keysize+name_hash_len:keysize+name_hash_len+10]
                    signature   = packet.data[keysize+name_hash_len+10:keysize+name_hash_len+10+sig_len]
                    app_data    = b""
                    if len(packet.data) > keysize+name_hash_len+10+sig_len:
                        app_data = packet.data[keysize+name_hash_len+10+sig_len:]

                signed_data = destination_hash+public_key+name_hash+random_hash+ratchet+app_data

                if not len(packet.data) > Identity.KEYSIZE//8+Identity.NAME_HASH_LENGTH//8+10+Identity.SIGLENGTH//8:
                    app_data = None

                announced_identity = Identity(create_keys=False)
                announced_identity.load_public_key(public_key)

                if len(RNS.Transport.blackholed_identities) > 0:
                    if announced_identity.hash in RNS.Transport.blackholed_identities:
                        RNS.log(f"Invalidated and dropped announce from blackholed identity {RNS.prettyhexrep(announced_identity.hash)}", RNS.LOG_EXTREME) if RNS.sl(RNS.LOG_EXTREME) else None
                        return False

                if announced_identity.pub != None and announced_identity.validate(signature, signed_data):
                    if only_validate_signature:
                        del announced_identity
                        return True

                    hash_material = name_hash+announced_identity.hash
                    expected_hash = RNS.Identity.full_hash(hash_material)[:RNS.Reticulum.TRUNCATED_HASHLENGTH//8]

                    if destination_hash == expected_hash:
                        # Check if we already have a public key for this destination
                        # and make sure the public key is not different.
                        if destination_hash in Identity.known_destinations:
                            if public_key != Identity.known_destinations[destination_hash][2]:
                                # In reality, this should never occur, but in the odd case
                                # that someone manages a hash collision, we reject the announce.
                                RNS.log("Received announce with valid signature and destination hash, but announced public key does not match already known public key.", RNS.LOG_CRITICAL)
                                RNS.log("This may indicate an attempt to modify network paths, or a random hash collision. The announce was rejected.", RNS.LOG_CRITICAL)
                                return False

                        RNS.Identity.remember(packet.get_hash(), destination_hash, public_key, app_data)
                        del announced_identity

                        if packet.rssi != None or packet.snr != None:
                            signal_str = " ["
                            if packet.rssi != None:
                                signal_str += "RSSI "+str(packet.rssi)+"dBm"
                                if packet.snr != None:
                                    signal_str += ", "
                            if packet.snr != None:
                                signal_str += "SNR "+str(packet.snr)+"dB"
                            signal_str += "]"
                        else:
                            signal_str = ""

                        if hasattr(packet, "transport_id") and packet.transport_id != None:
                            RNS.log("Valid announce for "+RNS.prettyhexrep(destination_hash)+" "+str(packet.hops)+" hops away, received via "+RNS.prettyhexrep(packet.transport_id)+" on "+str(packet.receiving_interface)+signal_str, RNS.LOG_EXTREME) if RNS.sl(RNS.LOG_EXTREME) else None
                        else:
                            RNS.log("Valid announce for "+RNS.prettyhexrep(destination_hash)+" "+str(packet.hops)+" hops away, received on "+str(packet.receiving_interface)+signal_str, RNS.LOG_EXTREME) if RNS.sl(RNS.LOG_EXTREME) else None

                        if ratchet:
                            Identity._remember_ratchet(destination_hash, ratchet)

                        return True

                    else:
                        RNS.log("Received invalid announce for "+RNS.prettyhexrep(destination_hash)+": Destination mismatch.", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
                        return False

                else:
                    RNS.log("Received invalid announce for "+RNS.prettyhexrep(destination_hash)+": Invalid signature.", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
                    del announced_identity
                    return False
        
        except Exception as e:
            RNS.log("Error occurred while validating announce. The contained exception was: "+str(e), RNS.LOG_ERROR)
            return False

    @staticmethod
    def persist_data(background=False):
        if not RNS.Transport.owner.is_connected_to_shared_instance:
            Identity.save_known_destinations(background=background)

    @staticmethod
    def exit_handler():
        Identity.persist_data()


    @staticmethod
    def from_bytes(prv_bytes):
        """
        Create a new :ref:`RNS.Identity<api-identity>` instance from *bytes* of private key.
        Can be used to load previously created and saved identities into Reticulum.

        :param prv_bytes: The *bytes* of private a saved private key. **HAZARD!** Never use this to generate a new key by feeding random data in prv_bytes.
        :returns: A :ref:`RNS.Identity<api-identity>` instance, or *None* if the *bytes* data was invalid.
        """
        identity = Identity(create_keys=False)
        if identity.load_private_key(prv_bytes):
            return identity
        else:
            return None


    @staticmethod
    def from_file(path):
        """
        Create a new :ref:`RNS.Identity<api-identity>` instance from a file.
        Can be used to load previously created and saved identities into Reticulum.

        :param path: The full path to the saved :ref:`RNS.Identity<api-identity>` data
        :returns: A :ref:`RNS.Identity<api-identity>` instance, or *None* if the loaded data was invalid.
        """
        identity = Identity(create_keys=False)
        if identity.load(path):
            return identity
        else:
            return None

    def to_file(self, path):
        """
        Saves the identity to a file. This will write the private key to disk,
        and anyone with access to this file will be able to decrypt all
        communication for the identity. Be very careful with this method.

        :param path: The full path specifying where to save the identity.
        :returns: True if the file was saved, otherwise False.
        """
        try:
            with open(path, "wb") as key_file:
                key_file.write(self.get_private_key())
                return True
            return False
        except Exception as e:
            RNS.log("Error while saving identity to "+str(path), RNS.LOG_ERROR)
            RNS.log("The contained exception was: "+str(e))

    def pub_to_file(self, path):
        """
        Saves the public identity to a file.

        :param path: The full path specifying where to save the identity.
        :returns: True if the file was saved, otherwise False.
        """
        try:
            with open(path, "wb") as key_file:
                key_file.write(self.get_public_key())
                return True
            return False
        except Exception as e:
            RNS.log("Error while saving identity to "+str(path), RNS.LOG_ERROR)
            RNS.log("The contained exception was: "+str(e))

    def __init__(self,create_keys=True):
        # Initialize keys to none
        self.prv           = None
        self.prv_bytes     = None
        self.sig_prv       = None
        self.sig_prv_bytes = None

        self.pub           = None
        self.pub_bytes     = None
        self.sig_pub       = None
        self.sig_pub_bytes = None

        self.hash          = None
        self.hexhash       = None

        if create_keys:
            self.create_keys()

    def create_keys(self):
        self.prv           = X25519PrivateKey.generate()
        self.prv_bytes     = self.prv.private_bytes()

        self.sig_prv       = Ed25519PrivateKey.generate()
        self.sig_prv_bytes = self.sig_prv.private_bytes()

        self.pub           = self.prv.public_key()
        self.pub_bytes     = self.pub.public_bytes()

        self.sig_pub       = self.sig_prv.public_key()
        self.sig_pub_bytes = self.sig_pub.public_bytes()

        self.update_hashes()

        RNS.log("Identity keys created for "+RNS.prettyhexrep(self.hash), RNS.LOG_VERBOSE)

    def get_private_key(self):
        """
        :returns: The private key as *bytes*
        """
        if self.prv_bytes and self.sig_prv_bytes: return self.prv_bytes+self.sig_prv_bytes
        else:                                     return None

    def get_public_key(self):
        """
        :returns: The public key as *bytes*
        """
        if self.pub_bytes and self.sig_pub_bytes: return self.pub_bytes+self.sig_pub_bytes
        else:                                     return None

    def load_private_key(self, prv_bytes):
        """
        Load a private key into the instance.

        :param prv_bytes: The private key as *bytes*.
        :returns: True if the key was loaded, otherwise False.
        """
        try:
            self.prv_bytes     = prv_bytes[:Identity.KEYSIZE//8//2]
            self.prv           = X25519PrivateKey.from_private_bytes(self.prv_bytes)
            self.sig_prv_bytes = prv_bytes[Identity.KEYSIZE//8//2:]
            self.sig_prv       = Ed25519PrivateKey.from_private_bytes(self.sig_prv_bytes)
            
            self.pub           = self.prv.public_key()
            self.pub_bytes     = self.pub.public_bytes()

            self.sig_pub       = self.sig_prv.public_key()
            self.sig_pub_bytes = self.sig_pub.public_bytes()

            self.update_hashes()

            return True

        except Exception as e:
            raise e
            RNS.log("Failed to load identity key", RNS.LOG_ERROR)
            RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)
            return False

    def load_public_key(self, pub_bytes):
        """
        Load a public key into the instance.

        :param pub_bytes: The public key as *bytes*.
        :returns: True if the key was loaded, otherwise False.
        """
        try:
            self.pub_bytes     = pub_bytes[:Identity.KEYSIZE//8//2]
            self.sig_pub_bytes = pub_bytes[Identity.KEYSIZE//8//2:]

            self.pub           = X25519PublicKey.from_public_bytes(self.pub_bytes)
            self.sig_pub       = Ed25519PublicKey.from_public_bytes(self.sig_pub_bytes)

            self.update_hashes()
        except Exception as e:
            RNS.log("Error while loading public key, the contained exception was: "+str(e), RNS.LOG_ERROR)

    def update_hashes(self):
        self.hash = Identity.truncated_hash(self.get_public_key())
        self.hexhash = self.hash.hex()

    def load(self, path):
        try:
            with open(path, "rb") as key_file:
                prv_bytes = key_file.read()
                return self.load_private_key(prv_bytes)
            return False
        except Exception as e:
            RNS.log("Error while loading identity from "+str(path), RNS.LOG_ERROR)
            RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)

    def get_salt(self):
        return self.hash

    def get_context(self):
        return None

    def encrypt(self, plaintext, ratchet=None):
        """
        Encrypts information for the identity.

        :param plaintext: The plaintext to be encrypted as *bytes*.
        :returns: Ciphertext token as *bytes*.
        :raises: *KeyError* if the instance does not hold a public key.
        """
        if self.pub != None:
            ephemeral_key = X25519PrivateKey.generate()
            ephemeral_pub_bytes = ephemeral_key.public_key().public_bytes()

            if ratchet != None:
                target_public_key = X25519PublicKey.from_public_bytes(ratchet)
            else:
                target_public_key = self.pub

            shared_key = ephemeral_key.exchange(target_public_key)
            
            derived_key = RNS.Cryptography.hkdf(
                length=Identity.DERIVED_KEY_LENGTH,
                derive_from=shared_key,
                salt=self.get_salt(),
                context=self.get_context(),
            )

            token = Token(derived_key)
            ciphertext = token.encrypt(plaintext)
            token = ephemeral_pub_bytes+ciphertext

            return token
        else:
            raise KeyError("Encryption failed because identity does not hold a public key")

    def __decrypt(self, shared_key, ciphertext):
        derived_key = RNS.Cryptography.hkdf(
            length=Identity.DERIVED_KEY_LENGTH,
            derive_from=shared_key,
            salt=self.get_salt(),
            context=self.get_context())

        token = Token(derived_key)
        plaintext = token.decrypt(ciphertext)
        return plaintext

    def decrypt(self, ciphertext_token, ratchets=None, enforce_ratchets=False, ratchet_id_receiver=None):
        """
        Decrypts information for the identity.

        :param ciphertext: The ciphertext to be decrypted as *bytes*.
        :returns: Plaintext as *bytes*, or *None* if decryption fails.
        :raises: *KeyError* if the instance does not hold a private key.
        """

        if self.prv != None:
            if len(ciphertext_token) > Identity.KEYSIZE//8//2:
                plaintext = None
                try:
                    peer_pub_bytes = ciphertext_token[:Identity.KEYSIZE//8//2]
                    peer_pub = X25519PublicKey.from_public_bytes(peer_pub_bytes)
                    ciphertext = ciphertext_token[Identity.KEYSIZE//8//2:]

                    if ratchets:
                        for ratchet in ratchets:
                            try:
                                ratchet_prv = X25519PrivateKey.from_private_bytes(ratchet)
                                ratchet_id = Identity._get_ratchet_id(ratchet_prv.public_key().public_bytes())
                                shared_key = ratchet_prv.exchange(peer_pub)
                                plaintext = self.__decrypt(shared_key, ciphertext)
                                if ratchet_id_receiver:
                                    ratchet_id_receiver.latest_ratchet_id = ratchet_id
                                
                                break
                            
                            except Exception as e:
                                pass

                    if enforce_ratchets and plaintext == None:
                        RNS.log("Decryption with ratchet enforcement by "+RNS.prettyhexrep(self.hash)+" failed. Dropping packet.", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
                        if ratchet_id_receiver:
                            ratchet_id_receiver.latest_ratchet_id = None
                        return None

                    if plaintext == None:
                        shared_key = self.prv.exchange(peer_pub)
                        plaintext = self.__decrypt(shared_key, ciphertext)

                        if ratchet_id_receiver:
                            ratchet_id_receiver.latest_ratchet_id = None

                except Exception as e:
                    RNS.log("Decryption by "+RNS.prettyhexrep(self.hash)+" failed: "+str(e), RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
                    if ratchet_id_receiver:
                        ratchet_id_receiver.latest_ratchet_id = None
                    
                return plaintext
            
            else:
                RNS.log("Decryption failed because the token size was invalid.", RNS.LOG_DEBUG) if RNS.sl(RNS.LOG_DEBUG) else None
                return None
        else:
            raise KeyError("Decryption failed because identity does not hold a private key")


    def sign(self, message):
        """
        Signs information by the identity.

        :param message: The message to be signed as *bytes*.
        :returns: Signature as *bytes*.
        :raises: *KeyError* if the instance does not hold a private key.
        """
        if self.sig_prv != None:
            try:
                return self.sig_prv.sign(message)    
            except Exception as e:
                RNS.log("The identity "+str(self)+" could not sign the requested message. The contained exception was: "+str(e), RNS.LOG_ERROR)
                raise e
        else:
            raise KeyError("Signing failed because identity does not hold a private key")

    def validate(self, signature, message):
        """
        Validates the signature of a signed message.

        :param signature: The signature to be validated as *bytes*.
        :param message: The message to be validated as *bytes*.
        :returns: True if the signature is valid, otherwise False.
        :raises: *KeyError* if the instance does not hold a public key.
        """
        if self.pub != None:
            try:
                self.sig_pub.verify(signature, message)
                return True
            except Exception as e:
                return False
        else:
            raise KeyError("Signature validation failed because identity does not hold a public key")

    def prove(self, packet, destination=None):
        signature = self.sign(packet.packet_hash)
        if RNS.Reticulum.should_use_implicit_proof():
            proof_data = signature
        else:
            proof_data = packet.packet_hash + signature
        
        if destination == None:
            destination = packet.generate_proof_destination()

        proof = RNS.Packet(destination, proof_data, RNS.Packet.PROOF, attached_interface = packet.receiving_interface)
        proof.send()

    def __str__(self):
        return RNS.prettyhexrep(self.hash)



#         Based on the original rnsh program by Aaron Heise (@acehoss)
# https://github.com/acehoss/rnsh - MIT License - Copyright (c) 2023 Aaron Heise
#     This version of rnsh is included in RNS under the Reticulum License
#
# Reticulum License
#
# Copyright (c) 2016-2026 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ._version import __version__

import os
module_abs_filename = os.path.abspath(__file__)
module_dir = os.path.dirname(module_abs_filename)

def _get_version(): return __version__



# Reticulum License
#
# Copyright (c) 2016-2025 Mark Qvist
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# - The Software shall not be used in any kind of system which includes amongst
#   its functions the ability to purposefully do harm to human beings.
#
# - The Software shall not be used, directly or indirectly, in the creation of
#   an artificial intelligence, machine learning or language model training
#   dataset, including but not limited to any use that contributes to the
#   training or development of such a model or algorithm.
#
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

def get_platform():
    from os import environ
    if "ANDROID_ARGUMENT" in environ: return "android"
    elif "ANDROID_ROOT" in environ:   return "android"
    else:
        import sys
        return sys.platform

def is_linux():
    if get_platform() == "linux": return True
    else: return False

def is_darwin():
    if get_platform() == "darwin": return True
    else: return False

def is_android():
    if get_platform() == "android": return True
    else: return False

def is_windows():
    if str(get_platform()).startswith("win"): return True
    else: return False

def use_epoll():
    if is_linux() or is_android(): return True
    else: return False

def use_af_unix():
    if is_linux() or is_android(): return True
    else: return False

def platform_checks():
    if is_windows():
        import sys
        if sys.version_info.major >= 3 and sys.version_info.minor >= 8: pass
        else:
            import RNS
            RNS.log("On Windows, Reticulum requires Python 3.8 or higher.", RNS.LOG_ERROR)
            RNS.log("Please update Python to run Reticulum.", RNS.LOG_ERROR)
            RNS.panic()

def cryptography_old_api():
    import cryptography
    if cryptography.__version__ == "2.8": return True
    else: return False
